#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""FAL command for Canon HOME-BOX Settings."""

import os

from iobase.exception import (NotFoundError, InvalidArgError)

import ionas.request as request
from iobase.xmldata import XmlNode

import libfal.falcmd as falcmd
import libfal.util as util

from fal_info import ReqTypes
from fal_info import Results

import fal_config

import pal_info

from pkgaddon.chbox_config import (SYNC_SERVICE_NAME, REST_BASE_URL)


X_ROOT = "chbox"
X_SHARELIST = "sharelist"
X_SHARE = "share"
X_NAME = "name"
X_SHARESERVICE = "shareservice"
X_SYNC = "chboxsync"
X_STATUS = "status"
X_PORTALID = "portalid"
X_ACCOUNT = "loginid"
X_PASSWD = "passwd"
X_CONTENTLIST = "folderlist"
X_CONTENT = "folder"
X_CONTENTPATH = "path"
X_CONTENTDESC = "desc"
X_DOWNLOADINTERVAL = "downloadinterval"
X_UPLOADMAXFILESIZE = "uploadmaxfilesize"

X_ID = "id"

X_SHARE_NAME = X_SHARE + "/" + X_NAME
X_SHARE_SYNC = X_SHARE + "/" + X_SHARESERVICE + "/" + X_SYNC + "[@enable]"

X_SYNC_STATUS = X_SHARE_SYNC + "/" + X_STATUS
X_SYNC_PORTALID = X_SHARE_SYNC + "/" + X_PORTALID
X_SYNC_ACCOUNT = X_SHARE_SYNC + "/" + X_ACCOUNT
X_SYNC_PASSWD = X_SHARE_SYNC + "/" + X_PASSWD
X_SYNC_DOWNLOADINTERVAL = X_SHARE_SYNC + "/" + X_DOWNLOADINTERVAL
X_SYNC_UPLOADMAXFILESIZE = X_SHARE_SYNC + "/" + X_UPLOADMAXFILESIZE
X_SYNC_CONTENTNAME = X_SHARE_SYNC + "/" + X_CONTENT + "/" + X_CONTENTPATH

X_LIST_PORTALID = X_SYNC + "/" + X_PORTALID
X_LIST_ACCOUNT = X_SYNC + "/" + X_ACCOUNT
X_LIST_PASSWD = X_SYNC + "/" + X_PASSWD

X_TARGET_SHARE_BASE = 'sharelist/share[@id="%s"]'
X_SYNC_NODE = 'shareservice/chboxsync[@enable]'

STATUS_NOERROR = "noerror"
STATUS_FAIL = "fail"

# Add the service entry for the clone package
util.config.FUNC_TO_SERVICE_MAP[SYNC_SERVICE_NAME] = SYNC_SERVICE_NAME


class FalCHBox(falcmd.FalCmd):
    """FAL CHBox command implementation class."""

    #
    # Definitions
    #

    CONF_FILE = os.path.join(fal_config.XML_CONF_DIR, 'config-chbox.xml')

    #
    # Internal methods
    #

    def __is_cloud_sync_enabled(self, entry):
        """Check sync config."""

        try:
            return entry.get().get_node(X_SHARE_SYNC).is_enabled()
        except:
            return False

    def __get_share_node_from_entry(self, entry):
        """Get share node from entry."""

        try:
            return entry.get().get_node(X_SHARE)
        except:
            raise InvalidArgError("share node not found.")

    def __get_share_id_from_entry(self, entry):
        """Get share ID from entry."""

        try:
            share_node = self.__get_share_node_from_entry(entry)
            return share_node.get_attr_value(X_ID)
        except:
            pass

        try:
            share_name = share_node.get_value(X_NAME)
            return util.encode_share_id(share_name)
        except:
            raise InvalidArgError("share ID not found.")

    def __get_config_share_node(self, share_id, module_config):
        """Get target share node in config."""

        try:
            return module_config.get_node(X_TARGET_SHARE_BASE % share_id)
        except:
            raise InvalidArgError("Matching share not found in config.")

    def __get_config_share_node_index(self, share_id, module_config):
        """Find share node and return index number."""

        share_list = module_config.get_node(X_SHARELIST).get_node_list(X_SHARE)
        index = 0
        found = False

        for share in share_list:
            if share.get_attr_value(X_ID) == share_id:
                found = True
                break
            index += 1

        if not found:
            raise NotFoundError("Share ID not found.")

        return index

    def __get_endpoint_url(self):
        """Get endpoint URL."""

        return REST_BASE_URL

    def __get_proxy(self):
        """Get proxy node."""

        try:
            return util.get_proxy(self.__get_endpoint_url())
        except util.NoProxy:
            pass

        return None

    def __pal_request_query(self, node):
        """Send query request to pal."""

        return self.send_entry(
            request.Entry(ReqTypes.QUERY, node.get_clone()),
            pal_info.PAL_CHBOX_PATH)

    def __pal_request_list(self, node):
        """Send list request to pal."""

        return self.send_entry(
            request.Entry(ReqTypes.LIST, node),
            pal_info.PAL_CHBOX_PATH)

    def __pal_request_download(self, node):
        """Send download request to pal."""

        return self.send_entry(
            request.Entry(ReqTypes.DOWNLOAD, node),
            pal_info.PAL_CHBOX_PATH)

    #
    # Request handlers
    #

    def __add_handler(self, entry):
        """Handle add request."""

        # validation
        try:
            util.check_root_node_name(entry, X_ROOT)
            util.check_node_existence(entry, X_SHARE)
            util.check_node_value_existence(entry, X_SHARE_NAME)
            util.check_node_existence(entry, X_SHARE_SYNC)
            if self.__is_cloud_sync_enabled(entry):
                util.check_node_value_existence(entry, X_SYNC_PORTALID)
                util.check_node_value_existence(entry, X_SYNC_ACCOUNT)
                util.check_node_existence(entry, X_SYNC_PASSWD)
                util.check_node_value_existence(entry, X_SYNC_CONTENTNAME)
                util.check_node_value_existence(
                    entry, X_SYNC_DOWNLOADINTERVAL)
                util.check_node_value_existence(
                    entry, X_SYNC_UPLOADMAXFILESIZE)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        # set config
        dst_entry = entry.get_clone()
        module_config = util.load_config(self.CONF_FILE)

        try:
            module_config.get_node(X_SHARELIST)
        except:
            module_config.append_child(X_SHARELIST)

        dst_share_node = dst_entry.get().get_node(X_SHARE)
        dst_share_name = dst_share_node.get_value(X_NAME)
        dst_share_id = util.encode_share_id(dst_share_name)
        dst_share_node.set_attrs({X_ID: dst_share_id})

        try:
            idx = self.__get_config_share_node_index(
                dst_share_id, module_config)
            module_config.get_node(X_SHARELIST).delete_child(
                X_SHARE, index=idx)
        except NotFoundError:
            pass

        try:
            module_config.get_node(X_SHARELIST).append_child(
                dst_entry.get().get_node(X_SHARE))
            util.save_config(self.CONF_FILE, module_config)
        except:
            dst_entry.set_result(Results.ADD_ERROR)
            return dst_entry

        # event
        try:
            util.servicectrl_suspend([], [SYNC_SERVICE_NAME, ])
            util.servicectrl_resume([], [SYNC_SERVICE_NAME, ])
        except:
            util.servicectrl_resume([], [SYNC_SERVICE_NAME, ])

        dst_entry.set_result(Results.OK)
        return dst_entry

    def __modify_handler(self, entry):
        """Handle modify request."""

        # validation
        try:
            util.check_root_node_name(entry, X_ROOT)
            src_share_node = self.__get_share_node_from_entry(entry)
            src_share_id = self.__get_share_id_from_entry(entry)
            module_config = util.load_config(self.CONF_FILE)
            config_share_node = self.__get_config_share_node(
                src_share_id, module_config)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        # set config
        try:
            config_share_node.update(src_share_node)
            util.save_config(self.CONF_FILE, module_config)
        except:
            entry.set_result(Results.MODIFY_ERROR)
            return entry

        # event
        try:
            util.servicectrl_suspend([], [SYNC_SERVICE_NAME, ])
            util.servicectrl_resume([], [SYNC_SERVICE_NAME, ])
        except:
            util.servicectrl_resume([], [SYNC_SERVICE_NAME, ])

        entry.set_result(Results.OK)
        return entry

    def __delete_handler(self, entry):
        """Handle delete request."""

        # validation
        try:
            util.check_root_node_name(entry, X_ROOT)
            util.check_node_existence(entry, X_SHARE)
            src_share_id = self.__get_share_id_from_entry(entry)
            module_config = util.load_config(self.CONF_FILE)
            module_config.get_node(X_TARGET_SHARE_BASE % src_share_id)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        # set config
        try:
            idx = self.__get_config_share_node_index(
                src_share_id, module_config)
            module_config.get_node(X_SHARELIST).delete_child(
                X_SHARE, index=idx)
            util.save_config(self.CONF_FILE, module_config)
        except:
            entry.set_result(Results.DELETE_ERROR)
            return entry

        # event
        try:
            util.servicectrl_suspend([], [SYNC_SERVICE_NAME, ])
            util.servicectrl_resume([], [SYNC_SERVICE_NAME, ])
        except:
            util.servicectrl_resume([], [SYNC_SERVICE_NAME, ])

        entry.set_result(Results.OK)
        return entry

    def __query_handler(self, entry):
        """Handle query request."""

        # validation
        try:
            util.check_root_node_name(entry, X_ROOT)
            util.check_node_existence(entry, X_SHARE)
            src_share_id = self.__get_share_id_from_entry(entry)
            module_config = util.load_config(self.CONF_FILE)
            config_share_node = self.__get_config_share_node(
                src_share_id, module_config)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        # make response
        dst_entry = entry.get_clone()
        dst_root_node = dst_entry.get()
        dst_share_node = dst_root_node.get_node(X_SHARE)
        dst_share_node.update(config_share_node)
        dst_sync_node = dst_share_node.get_node(X_SYNC_NODE)
        if dst_sync_node.is_enabled():
            try:
                # get status
                response_entry = self.__pal_request_query(dst_root_node)
                if response_entry.get_result() == Results.OK:
                    dst_status = response_entry.get().get_value(X_SYNC_STATUS)
                    dst_sync_node.append_child(X_STATUS, text=dst_status)
                else:
                    dst_sync_node.append_child(X_STATUS, text=STATUS_FAIL)
            except:
                dst_sync_node.append_child(X_STATUS, text=STATUS_FAIL)

        dst_entry.set_result(Results.OK)
        return dst_entry

    def __list_handler(self, entry):
        """Handle list request."""

        # validation
        try:
            util.check_root_node_name(entry, X_ROOT)
            util.check_node_value_existence(entry, X_LIST_PORTALID)
            util.check_node_value_existence(entry, X_LIST_ACCOUNT)
            util.check_node_existence(entry, X_LIST_PASSWD)
            src_root_node = entry.get()
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        # proxy
        try:
            src_proxy_node = self.__get_proxy()
            if src_proxy_node:
                src_root_node.append_child(src_proxy_node)
        except:
            pass

        # send request to pal
        response_entry = self.__pal_request_list(src_root_node)
        if response_entry.get_result() != Results.OK:
            entry.set_result(Results.LIST_ERROR)
            return entry

        dst_entry = response_entry.get_clone()
        dst_entry.set_result(Results.OK)
        return dst_entry

    def __download_handler(self, entry):
        """Handle download request."""

        # validation
        try:
            util.check_root_node_name(entry, X_ROOT)
            util.check_node_existence(entry, X_SHARE)
            src_share_id = self.__get_share_id_from_entry(entry)
            module_config = util.load_config(self.CONF_FILE)
            config_share_node = self.__get_config_share_node(
                src_share_id, module_config)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        # send request to pal
        dst_root_node = XmlNode(X_ROOT)
        dst_root_node.append_child(config_share_node)
        response_entry = self.__pal_request_download(dst_root_node)
        if response_entry.get_result() != Results.OK:
            entry.set_result(Results.DOWNLOAD_ERROR)
            return entry

        dst_entry = response_entry.get_clone()
        dst_entry.set_result(Results.OK)
        return dst_entry

    #
    # Constructor
    #

    def __init__(self, option=None):
        """Constructor."""

        falcmd.FalCmd.__init__(self, option=option)

        load_save_handler = lambda entry: self.send_entry_with_file(
            entry, pal_info.PAL_CONFFILE_PATH, self.CONF_FILE)

        self.set_handler(ReqTypes.LOAD, load_save_handler)
        self.set_handler(ReqTypes.SAVE, load_save_handler)

        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.MODIFY, self.__modify_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.LIST, self.__list_handler)
        self.set_handler(ReqTypes.DOWNLOAD, self.__download_handler)


#
# Main
#

if __name__ == "__main__":
    FalCHBox(option=falcmd.FalCmdOption()).handle_request()
