#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta
from __future__ import with_statement

from ionas.fileaccess import (
    MODE_READ,
    MODE_WRITE,
    MODE_APPEND,
    DEFAULT_LOCALE,
)
from ionas.exception import *
from ionas.exception import SizeTooLargeError
from ionas.fileaccess.filelikeobj import FileLikeObj

import stat
import locale

import urllib2

from chbox.chboxaccess import ChboxAccess as HomeBox2Access


# RETRY_COUNT = 3



def _decode(s, encoding):
    return s if isinstance(s, unicode) else s.decode(encoding)


def to_open_flag(mode):
    if mode == MODE_READ:
        return "r"
    if mode == MODE_WRITE:
        return "w"
    if mode == MODE_APPEND:
        return "w"
    return "r"


class HomeBox2FileObject(FileLikeObj):

    def __init__(self, file, path, mode, accessor, length, max_size=0):

        FileLikeObj.__init__(self, file, path)
        self.mode = mode
        self.accessor = accessor
        self.length = length
        self._max_size = max_size

    def __len__(self):

        return int(self.accessor.stat(self.path)[stat.ST_SIZE])

    def read(self, size):
        return self.__read(size)

    def __read(self, size):
        self.point += size
        return self.file.read(size)

    def read_by_path(self, file_like):
        return self.__read_by_path(file_like)

    def __read_by_path(self, file_like):
        file_like.write_by_path(self.file)

    def write(self, file_like):
        if self.file is None:
            open_mode = to_open_flag(self.mode)
            try:
                self.length = file_like.get_stat()[stat.ST_SIZE]
            except Exception as err:
                self.length = len(file_like)
            # self.__renew_file()
            self.file = self.accessor.access.open(
                self.path,
                open_mode,
                # length=self.length,
            )
        return self.__write(file_like)

    def __write(self, file_like):
        self.file.write_by_path(file_like)

    def write_by_path(self, file_like):
        # TODO
        size_b = file_like.get_stat()[stat.ST_SIZE]
        max_size_b = self.max_size * 1024 * 1024
        if max_size_b and max_size_b < size_b:
            raise SizeTooLargeError
        return self.__write_by_path(file_like)

    def __write_by_path(self, file_like):
        self.file.write_by_path(file_like)

    def close(self):
        self.__close()

    def __close(self):
        self.file.close()

    def get_stat(self):
        return self.accessor.stat(self.path)

    @property
    def max_size(self):
        return self._max_size



class BackupHomeBox2Access(object):

    DEFAULT_ENCODING = locale.getdefaultlocale()[1] or DEFAULT_LOCALE
    SEPARATOR = '/'

    def __init__(self,
                 target_info,
                 get_proxy=lambda url: (None, None),
                 encoding=DEFAULT_ENCODING):

        self.sep = self.SEPARATOR  # TODO: 不要

        self.encoding = encoding

        self.target_info = target_info

        # TODO: REST_BASE_URL from self.chbox_info.base_url
        from pkgaddon.chbox_config import REST_BASE_URL
        proxy_host, proxy_port = get_proxy(REST_BASE_URL)

        self.__folder = self.__pathnormalize(target_info.folder)

        self.access = HomeBox2Access(
            target_info.portalid,
            target_info.loginid,
            target_info.password,
            encoding=self.encoding,
            proxy_host=proxy_host,
            proxy_port=proxy_port,
            is_session_open=True,
        )

    def listdir(self, path):
        # path = _decode(path, self.encoding)
        path = self.__pathnormalize(path)
        return [e.encode(self.encoding) for e in self.access.listdir(path)]

    def isdir(self, entry):
        raise Exception('isdir')  # FIXME

    def open(self, path, mode):
        # path = _decode(path, self.encoding)
        path = self.__pathnormalize(path)
        # open_mode = to_open_flag(mode)
        # opened_file = self.access.open(path, open_mode)
        opened_file = self.access.open(path, mode)
        return HomeBox2FileObject(
            opened_file,
            path,
            mode,
            accessor=self,
            length=-1,
            max_size=self.target_info.max_upload_size,
        )

    def stat(self, path):
        # path = _decode(path, self.encoding)
        path = self.__pathnormalize(path)
        return self.access.stat(path)

    def chmod(self, path, mode):
        return NotImplementedError

    def utime(self, path, mtime):
        return NotImplementedError

    def remove(self, path):
        # path = _decode(path, self.encoding)
        path = self.__pathnormalize(path)
        self.access.remove(path)

    def rmdir(self, path, noexist_ok=True):
        # path = _decode(path, self.encoding)
        path = self.__pathnormalize(path)
        try:
            self.access.rmdir(path)
        except urllib2.HTTPError as e:
            # 404: Not Found
            if e.code == 404 and noexist_ok:
                pass
            else:
                raise

    def mkdir(self, path, exist_ok=True, recursive=True):
        # _path = _decode(path, self.encoding)
        _path = path
        _path = self.__pathnormalize(_path)
        try:
            self.access.mkdir(_path)
        except urllib2.HTTPError as e:
            if   e.code == 404:  # 404: Not Found
                if recursive:
                    _head, _tail = self.__pathsplit(_path)
                    self.mkdir(_head, exist_ok=exist_ok)
                    self.mkdir(_path, exist_ok=exist_ok, recursive=False)
                else:
                    raise
            elif e.code == 409:  # 409: Conflict
                if exist_ok:
                    pass
                else:
                    raise
            else:
                if self.__exist(path):
                    pass
                else:
                    raise

    def move(self, src_path, dst_path):
        src_path = self.__pathnormalize(src_path)
        dst_path = self.__pathnormalize(dst_path)
        # TODO
        _src_head, _src_tail = self.__pathsplit(src_path)
        _dst_head, _dst_tail = self.__pathsplit(dst_path)
        self.access.file_client.rename_directory(src_path, _dst_tail)
        if _src_head != _dst_head:
            _src_path2 = self.__pathjoin(_src_head, _dst_tail)
            self.access.file_client.move_directory(_src_path2, _dst_head)

    def __exist(self, path):
        # TODO
        try:
            self.stat(path)
            return True
        except Exception as err:
            return False

    def __pathsplit(self, path):
        separator = self.sep
        names = path.split(separator)
        head = separator.join(names[:-1])
        tail = names[-1]
        return head, tail

    def __pathjoin(self, head, tail):
        separator = self.sep
        return separator.join([head, tail])

    def __pathnormalize(self, path):
        # TODO
        path = path.replace('{0}{0}'.format(self.sep), '{0}'.format(self.sep))
        path = path.rstrip(self.sep)
        return path
