#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Canon HOME-BOX Settings."""

import stat
import time
import urllib2
import httplib

from iobase.exception import InvalidArgError
from iobase.cmdexec import (CmdExec, NULL, PIPE)
from iobase.xmldata import XmlConfNode

from ionas.restapi import client

import libpal.palcmd as palcmd
import libpal.util as util

from pal_info import (ReqTypes, Results)

from chbox.chboxinfo import CHBoxInfo
from pkgaddon.chbox_config import (
    REST_BASE_URL, APPLICATION_ID, TARGET_PATH_LIST)
from pkgaddon.pal_chbox_config import CHBOXSYNC_CMD


X_SHARE = "share"
X_NAME = "name"
X_SHARESERVICE = "shareservice"
X_SYNC = "chboxsync"
X_STATUS = "status"
X_PORTALID = "portalid"
X_ACCOUNT = "loginid"
X_PASSWD = "passwd"
X_CONTENTLIST = "folderlist"
X_CONTENT = "folder"
X_CONTENTPATH = "path"
X_CONTENTDESC = "desc"

X_PROXY = "proxy"

X_SHARE_NAME = X_SHARE + "/" + X_NAME
X_SHARE_SYNC = X_SHARE + "/" + X_SHARESERVICE + "/" + \
               X_SYNC + "[@enable='true']"

K_STAT = "stat"
K_NAME = "name"
K_DESC = "description"

SYNC_CMD_OPTION_INFO = "--info"
SYNC_CMD_OPTION_FORCE_DOWNLOAD = "--force-download"

SYNC_CMD_OUTPUT_STATUS_NOERROR = "status: noerror"

MAX_RETRY = 1


class PalCHBox(palcmd.PalCmd):
    """Pal CHBox command implementation class."""

    #
    # Definitions
    #

    STATUS_NOERROR = "noerror"
    STATUS_FAIL = "fail"

    #
    # Internal methods
    #

    def __execute_cloud_sync_command(
            self, command, option, share_name, stdout):
        """Execute cloud sync command with specified option."""

        if (option != SYNC_CMD_OPTION_INFO and
            option != SYNC_CMD_OPTION_FORCE_DOWNLOAD):
            raise InvalidArgError("Invalid option argument.")

        cmdline = [command, option, share_name]
        return CmdExec(cmdline, stdout=stdout, stderr=NULL).execute()

    def __list_contents(self, session, target):
        """List contents in target with retry."""

        for count in range(0, MAX_RETRY):
            try:
                return session.list_contents(target)
            except urllib2.HTTPError:
                if count == MAX_RETRY - 1:
                    raise
                time.sleep(1)
            except:
                raise

    def __close_session(self, session):
        """Close session with retry."""

        for count in range(0, MAX_RETRY):
            try:
                session.close_session()
                break
            except urllib2.HTTPError:
                if count == MAX_RETRY - 1:
                    raise
                time.sleep(1)
            except:
                raise

    def __is_content_directory(self, content):
        """Check content is directory."""

        if not K_STAT in content:
            return False

        if not stat.ST_MODE in content[K_STAT]:
            return False

        if stat.S_ISDIR(content[K_STAT][stat.ST_MODE]):
            return True

        return False

    #
    # Request handlers
    #

    def __query_handler(self, entry):
        """Handle query request."""

        try:
            src_root_node = entry.get()
            src_share_name = src_root_node.get_value(X_SHARE_NAME)
            src_root_node.get_node(X_SHARE_SYNC)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        try:
            dst_entry = entry.get_clone()
            dst_root_node = dst_entry.get()
            dst_sync_node = dst_root_node.get_node(X_SHARE_SYNC)
            dst_sync_node.append_child(X_STATUS, text=self.STATUS_FAIL)
        except:
            dst_entry.set_result(Results.QUERY_ERROR)
            return dst_entry

        try:
            proc = self.__execute_cloud_sync_command(
                CHBOXSYNC_CMD, SYNC_CMD_OPTION_INFO,
                src_share_name, PIPE)
            for line in proc.stdout.readlines():
                if line.startswith(SYNC_CMD_OUTPUT_STATUS_NOERROR):
                    dst_status_node = dst_sync_node.get_node(X_STATUS)
                    dst_status_node.set_text(self.STATUS_NOERROR)
        except:
            dst_entry.set_result(Results.QUERY_ERROR)
            return dst_entry

        dst_entry.set_result(Results.OK)
        return dst_entry

    def __list_handler(self, entry):
        """Handle list request."""

        try:
            src_root_node = entry.get()
            src_sync_node = src_root_node.get_node(X_SYNC)
            src_cabinet = src_sync_node.get_value(X_PORTALID)
            src_account = src_sync_node.get_value(X_ACCOUNT)
            src_passwd = src_sync_node.get_value(X_PASSWD)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        proxies = None
        try:
            src_proxy_node = src_root_node.get_node(X_PROXY)
            proxies = util.get_proxies(src_proxy_node)
        except:
            pass

        try:
            dst_entry = entry.get_clone()
        except:
            dst_entry.set_result(Results.LIST_ERROR)
            return dst_entry

        session = None
        dst_content_list = []
        try:
            session = client.FileClient(
                CHBoxInfo(
                    REST_BASE_URL, src_cabinet, src_account, src_passwd,
                    APPLICATION_ID),
                proxies=proxies)
            session.open_session()

            for target in TARGET_PATH_LIST:
                try:
                    for content in self.__list_contents(session, target):
                        if self.__is_content_directory(content):
                            dst_content = {
                                K_NAME: target + u"/" + content[K_NAME],
                                K_DESC: None}
                            dst_content_list.append(dst_content)
                except urllib2.HTTPError, err:
                    if err.code != httplib.NOT_FOUND:
                        raise
        except:
            dst_entry.set_result(Results.LIST_ERROR)
            return dst_entry
        finally:
            try:
                self.__close_session(session)
            except:
                dst_entry.set_result(Results.LIST_ERROR)
                return dst_entry

        try:
            dst_root_node = dst_entry.get()
            dst_sync_node = dst_root_node.get_node(X_SYNC)

            dst_contentlist_node = XmlConfNode(X_CONTENTLIST)
            for dst_content in dst_content_list:
                try:
                    dst_content_node = XmlConfNode(X_CONTENT)
                    dst_content_node.append_child(
                        X_CONTENTPATH, text=dst_content[K_NAME])
                    dst_content_node.append_child(
                        X_CONTENTDESC, text=dst_content[K_DESC])
                    dst_contentlist_node.append_child(dst_content_node)
                except:
                    pass
            dst_sync_node.append_child(dst_contentlist_node)
        except:
            dst_entry.set_result(Results.LIST_ERROR)
            return dst_entry

        dst_entry.set_result(Results.OK)
        return dst_entry

    def __download_handler(self, entry):
        """Handle download request."""

        try:
            src_root_node = entry.get()
            src_share_name = src_root_node.get_value(X_SHARE_NAME)
            src_root_node.get_node(X_SHARE_SYNC)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        try:
            dst_entry = entry.get_clone()
        except:
            dst_entry.set_result(Results.DOWNLOAD_ERROR)
            return dst_entry

        try:
            self.__execute_cloud_sync_command(
                CHBOXSYNC_CMD, SYNC_CMD_OPTION_FORCE_DOWNLOAD,
                src_share_name, NULL)
        except:
            dst_entry.set_result(Results.DOWNLOAD_ERROR)
            return dst_entry

        dst_entry.set_result(Results.OK)
        return dst_entry

    #
    # Constructor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.LIST, self.__list_handler)
        self.set_handler(ReqTypes.DOWNLOAD, self.__download_handler)


#
# Main
#

if __name__ == "__main__":
    PalCHBox(option=palcmd.PalCmdOption()).handle_request()
