#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Event Handler for unsetting HOME-BOX."""

import ionas.request as request

import libevh.evhcmd as evhcmd
import libevh.util as util

import fal_info

from product_conf import STOR_IF_TO_VOL
from product_conf.reqinfo import Results, ReqTypes


X_EVENTINFO = "eventinfo"
X_SHARELIST_SHARE = "sharelist/share"
X_SHARES_SHARE = "shares/share"
X_NAME = "name"
X_VOLUME = "volume"

A_CATEGORY = "category"
A_ACTION = "action"

CATEGORY_VOLUME = "volume"
CATEGORY_SHARES = "shares"

ACTION_MKFS = "mkfs"
ACTION_DELETE = "delete"

IF_INTERNAL = "internal"

TARGET_DELETE_FMT = "<chbox>%s</chbox>"


class EvhCHBoxUnset(evhcmd.EvhCmd):
    """Event Handler class for unsetting HOME-BOX."""

    #
    # Utility method
    #

    def __send_request(self, request_type, target_module, child_node=""):
        """Send request to target module."""

        entry = request.Entry(request_type, child_node)
        return self.send_entry(entry, target_module)

    def __load_shares_config(self):
        """Load config of shares."""

        response_entry = self.__send_request(
            ReqTypes.LOAD, fal_info.FAL_SHARES_PATH, "")
        if response_entry.get_result() != Results.OK:
            return None

        return response_entry.get()

    def __load_target_config(self):
        """Load config of HOME-BOX."""

        response_entry = self.__send_request(
            ReqTypes.LOAD, fal_info.FAL_CHBOX_PATH, "")
        if response_entry.get_result() != Results.OK:
            return None

        return response_entry.get()

    def __delete_target_share(self, target_share):
        """Delete target share."""

        request_string = TARGET_DELETE_FMT % str(target_share)
        response_entry = self.__send_request(
            ReqTypes.DELETE, fal_info.FAL_CHBOX_PATH, request_string)
        if response_entry.get_result() != Results.OK:
            return None

        return response_entry.get()

    def __is_volume_mkfs_event(self, eventinfo_node):
        """Check if event is volume mkfs."""

        category = eventinfo_node.get_attr_value(A_CATEGORY)
        action = eventinfo_node.get_attr_value(A_ACTION)

        if category != CATEGORY_VOLUME:
            return False
        if action != ACTION_MKFS:
            return False

        return True

    def __is_shares_delete_event(self, eventinfo_node):
        """Check if event is shares delete."""

        category = eventinfo_node.get_attr_value(A_CATEGORY)
        action = eventinfo_node.get_attr_value(A_ACTION)

        if category != CATEGORY_SHARES:
            return False
        if action != ACTION_DELETE:
            return False

        return True

    def __is_valid_event(self, entry):
        """Check if event is valid."""

        try:
            events_node = entry.get()
            eventinfo_node = events_node.get_node(X_EVENTINFO)
            if self.__is_volume_mkfs_event(eventinfo_node):
                if util.is_internal_vol_name(entry):
                    return True
            elif self.__is_shares_delete_event(eventinfo_node):
                return True
        except:
            pass

        return False

    def __get_shares_share_list(self, entry):
        """Get unset target shares."""

        share_list = []
        events_node = entry.get()
        eventinfo_node = events_node.get_node(X_EVENTINFO)

        if self.__is_volume_mkfs_event(eventinfo_node):
            shares_config = self.__load_shares_config()
            share_list = shares_config.get_node_list(X_SHARELIST_SHARE)
        elif self.__is_shares_delete_event(eventinfo_node):
            share_list.append(eventinfo_node.get_node(X_SHARES_SHARE))

        return share_list

    def __is_deletion_target_share(self, target_share, compare_share_list):
        """Check share is deletion target."""

        target_name = target_share.get_value(X_NAME)
        for compare_share in compare_share_list:
            compare_name = compare_share.get_value(X_NAME)
            if compare_name != target_name:
                continue

            compare_volume = compare_share.get_value(X_VOLUME)
            if not compare_volume:
                # 共有削除イベントの場合は常に削除対象
                return True
            elif compare_volume in STOR_IF_TO_VOL[IF_INTERNAL]:
                return True
            else:
                return False

        return False

    def __try_unset(self, entry):
        """Unset HOME-BOX."""

        try:
            target_config = self.__load_target_config()
            target_share_list = target_config.get_node_list(X_SHARELIST_SHARE)
            if not target_share_list:
                return True
        except:
            return False

        try:
            shares_share_list = self.__get_shares_share_list(entry)
            for target_share in target_share_list:
                is_target = self.__is_deletion_target_share(
                    target_share, shares_share_list)
                if not is_target:
                    continue

                self.__delete_target_share(target_share)
        except:
            return False

        return True

    #
    # Request handlers
    #

    def __event_handler(self, entry):
        """Handle "event" request."""

        if self.__is_valid_event(entry):
            self.__try_unset(entry)

        entry.set_result(Results.OK)
        return entry

    #
    # Constructor
    #

    def __init__(self):
        """Constructor."""

        evhcmd.EvhCmd.__init__(self, evhcmd.EvhCmdOption())
        self.set_handler(ReqTypes.EVENT, self.__event_handler)

#
# Main
#

if __name__ == "__main__":
    EvhCHBoxUnset().handle_request()
