#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from __future__ import with_statement

from ionas.fileaccess import (MODE_READ, MODE_WRITE, DEFAULT_LOCALE)
from ionas.exception import *
from iobase.exception import (InvalidArgError)
from ionas.fileaccess.filelikeobj import FileLikeObj

import os
import sys
import stat
import locale
import errno
import chbox.chboxinfo as chboxinfo
from ionas.restapi import client
from pkgaddon.chbox_config import REST_BASE_URL

APPLICATION_ID = "GLQWM"

RETRY_MAX = 1  # 現行はリトライしない
retry_interval = 3


##
# コンテナオブジェクト
#
# データを保持するだけのオブジェクト．何もしない．
#
class Container:
    pass

PRIORITY_NONE = "none"
PRIORITY_FILE = "file"
PRIORITY_DIR = "dir"

LOCAL_CANCEL_ERRS = (errno.EACCES, errno.EPERM,
                     errno.ENOENT, errno.EEXIST,
                     errno.ENAMETOOLONG)

CHBOX_SEP = "/"


def default_except_handler(err):
    raise err


class ChboxFileObj(FileLikeObj):

    def __init__(self, file, path, file_name, mode, access):
        self.file_name = file_name
        self.readobj = None
        self.access = access
        self.mode = mode

        if self.mode == MODE_WRITE:
            self.filesize = 0
        elif self.mode == MODE_READ:
            self.filesize = -1  # < 0 で未取得を表す
        else:
            raise InvalidArgError("invalid access mode %s" % str(mode))

        FileLikeObj.__init__(self, file, path)

    def __len__(self):
        if self.filesize >= 0:
            return self.filesize

        err = None
        for r in range(0, RETRY_MAX):
            try:
                self.filesize = self.access.stat(self.path)[stat.ST_SIZE]
                return self.filesize
            except:
                cl, err = sys.exc_info()[:2]
                self.access.except_handler(err)
        else:
            self.access.except_handler(err)

    def tell(self):
        raise NotImplementedError

    def seek(self, offset, whence=os.SEEK_SET):
        raise NotImplementedError

    def get_path(self):

        raise NotImplementedError

    def read(self, size=None):
        if self.readobj:
            return self.readobj.read(size)

        err = None
        for r in range(0, RETRY_MAX):
            try:
                self.readobj = self.file.read_file(self.path)
            except:
                cl, err = sys.exc_info()[:2]
                self.access.except_handler(err)
            else:
                # 正常終了
                return self.readobj.read(size)

        self.access.except_handler(err)

    def __update_file_size(self, file_like):

        try:
            self.filesize = len(file_like)
        except:
            try:
                self.filesize = os.fstat(file_like.fileno()).st_size
            except:
                pass

    def __write_by_path_with_mkdir(self, dir_path, file_like):

        try:
            self.access._mkdir_recursive(dir_path)
            self.file.write_file(self.path, file_like, overwrite=True)
        except:
            cl, err = sys.exc_info()[:2]
            self.access.except_handler(err)
        else:
            self.__update_file_size(file_like)
        return

    def write_by_path(self, file_like):

        path_dir, file_name = \
            self.path.rstrip(self.access.sep).rsplit(self.access.sep, 1)

        err = None
        for r in range(0, RETRY_MAX):
            try:
                self.file.write_file(self.path, file_like, overwrite=True)
            except:
                cl, err = sys.exc_info()[:2]
                try:
                    self.access.except_handler(err)
                except (NoEntryError, ExistsError):
                    if self.access.isdir(self.path, is_need_conv_path=False):
                        self.access.remove(self.path, maybe_dir=True,
                                           is_need_conv_path=False)
                    file_like.seek(0)
                    self.__write_by_path_with_mkdir(path_dir, file_like)
                    return
            else:
                self.__update_file_size(file_like)
                return

        self.access.except_handler(err)

    def close(self):
        pass


class ChboxAccess():

    def __init__(self,
                 portalid,
                 login_id,
                 login_passwd,
                 encoding=None,
                 proxy_host=None,
                 proxy_port=None,
                 except_handler=None,
                 is_session_open=True):

        self.sep = CHBOX_SEP  # パス区切り文字

        if except_handler:
            self.except_handler = except_handler
        else:
            self.except_handler = default_except_handler

        if encoding:
            self.encoding = encoding
        else:
            self.encoding = locale.getdefaultlocale()[1]
            if not self.encoding:
                self.encoding = DEFAULT_LOCALE

        # パラメータをインスタンス変数に保存
        self.portalid = portalid
        self.login_id = login_id
        self.login_passwd = login_passwd
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.file_client = None

        self.chbox_info = chboxinfo.CHBoxInfo(REST_BASE_URL,
                                              portalid,
                                              login_id,
                                              login_passwd,
                                              APPLICATION_ID)

        if not is_session_open:
            return

        err = None
        for r in range(0, RETRY_MAX):
            try:
                self.file_client = \
                    client.FileClient(self.chbox_info,
                                      proxies=self.__get_proxys())
                self.file_client.open_session()

            except:
                cl, err = sys.exc_info()[:2]
                try:
                    self.except_handler(err)
                finally:
                    try:
                        self.logout()
                    except:
                        pass
            else:
                break
        else:
            # リトライ回数オーバー
            try:
                self.logout()
            except:
                pass
            raise err

    def __get_server_path(self, path):

        try:
            _path = path.decode(self.encoding)
            return self.chbox_info.convert_path(_path)
        except:
            cl, err = sys.exc_info()[:2]
            self.except_handler(err)

    def __get_proxys(self):

        if self.proxy_host and self.proxy_port:
            proxys = {
                "http": "http://%s:%d/" % \
                            (self.proxy_host, int(self.proxy_port)),
                "https": "http://%s:%d/" % \
                             (self.proxy_host, int(self.proxy_port)),
            }
            return proxys
        return None

    def is_session_opened(self):

        if self.file_client:
            return True
        return False

    def session_open(self):

        if self.is_session_opened():
            return

        try:
            self.file_client = \
                client.FileClient(self.chbox_info,
                                  proxies=self.__get_proxys())
        except:
            self.file_client = None
            raise

        err = None
        for r in range(0, RETRY_MAX):
            try:
                self.file_client.open_session()
            except:
                cl, err = sys.exc_info()[:2]
                self.except_handler(err)
            else:
                break
        else:
            self.except_handler(err)

    def logout(self):

        try:
            if self.file_client:
                self.file_client.close_session()
                self.file_client = None
        except:
            pass

    def __del__(self):
        try:
            self.logout()
        except:
            pass

    def listdir(self, path):

        self.session_open()
        # _path = self.__conv_to_internal_path(path)
        _path = self.__get_server_path(path)

        try:
            contents_list = self.file_client.list_contents(_path)
        except:
            cl, err = sys.exc_info()[:2]
            self.except_handler(err)

        return [c["name"] for c in contents_list]

    def isdir(self, path, is_need_conv_path=True):

        if is_need_conv_path:
            _path = self.__get_server_path(path)
        else:
            _path = path

        self.session_open()

        try:
            stat_info = self.stat(_path, is_need_conv_path=False)
            return stat.S_ISDIR(stat_info[stat.ST_MODE])
        except:
            return False

    def isfile(self, path, is_need_conv_path=True):

        if is_need_conv_path:
            _path = self.__get_server_path(path)
        else:
            _path = path

        self.session_open()

        try:
            stat_info = self.stat(_path, is_need_conv_path=False)
            return stat.S_ISREG(stat_info[stat.ST_MODE])
        except:
            return False

    def open(self, path, mode):

        self.session_open()

        _path = self.__get_server_path(path)
        path_list = _path.rsplit(self.sep, 1)
        filename = path_list[-1]
        return ChboxFileObj(self.file_client, _path, filename, mode, self)

    def stat(self, path, maybe_dir=False, is_need_conv_path=True):

        self.session_open()

        if is_need_conv_path:
            _path = self.__get_server_path(path)
        else:
            _path = path

        stat_info = {}
        err = None
        for r in range(0, RETRY_MAX):
            try:
                if maybe_dir:
                    try:
                        stat_info = self.file_client.stat_directory(_path)
                    except:
                        stat_info = self.file_client.stat_file(_path)
                else:
                    try:
                        stat_info = self.file_client.stat_file(_path)
                    except:
                        stat_info = self.file_client.stat_directory(_path)
            except:
                cl, err = sys.exc_info()[:2]
                self.except_handler(err)
            else:
                # 正常終了
                break
        else:
            # リトライ回数オーバー
            self.except_handler(err)

        return stat_info

    def remove(self, path, maybe_dir=False, is_need_conv_path=True):

        self.session_open()

        if is_need_conv_path:
            _path = self.__get_server_path(path)
        else:
            _path = path

        err = None
        for r in range(0, RETRY_MAX):
            try:
                if maybe_dir:
                    try:
                        self.file_client.delete_directory(_path)
                    except:
                        self.file_client.delete_file(_path)
                else:
                    try:
                        self.file_client.delete_file(_path)
                    except:
                        self.file_client.delete_directory(_path)

            except:
                cl, err = sys.exc_info()[:2]
                self.except_handler(err)
            else:
                # 正常終了
                break
        else:
            # リトライ回数オーバー
            self.except_handler(err)

    def rmdir(self, path):
        self.remove(path, maybe_dir=True)

    def __is_exist_dir(self, path):

        try:
            self.file_client.stat_directory(path)
        except:
            cl, err = sys.exc_info()[:2]
            try:
                self.except_handler(err)
            except NoEntryError:
                return False
        return True

    def _mkdir_recursive(self, path):

        path_list = path.split(self.sep)

        # 上から順番に作る
        for idx in range(len(path_list) + 1):

            _path = self.sep.join(path_list[:idx])
            try:
                if self.sep == _path or not _path:
                    continue
                if not self.__is_exist_dir(_path):
                    self.file_client.make_directory(_path)
            except:
                cl, err = sys.exc_info()[:2]
                try:
                    self.except_handler(err)
                except ExistsError:
                    continue

    def mkdir(self, path):

        self.session_open()

        _path = self.__get_server_path(path)

        err = None
        for r in range(0, RETRY_MAX):
            try:
                self.file_client.make_directory(_path)
            except:
                cl, err = sys.exc_info()[:2]
                try:
                    self.except_handler(err)
                except (NoEntryError, ExistsError):
                    if self.isfile(_path, is_need_conv_path=False):
                        self.remove(_path, maybe_dir=False,
                                    is_need_conv_path=False)
                    self._mkdir_recursive(_path)
                    return
            else:
                # 正常終了
                break
        else:
            self.except_handler(err)

    def _make_dir_item(self, path, contents):
        c = Container()
        c.name = (path + self.sep + contents["name"] + self.sep).encode(
                  self.encoding)
        c.last_modified = contents["stat"][stat.ST_MTIME]
        c.size = contents["stat"][stat.ST_SIZE]
        return c

    def _make_file_item(self, path, contents):
        c = Container()
        c.name = (path + self.sep + contents["name"]).encode(self.encoding)
        c.last_modified = contents["stat"][stat.ST_MTIME]
        c.size = contents["stat"][stat.ST_SIZE]
        return c

    def _get_list(self, path):

        self.session_open()

        err = None
        contents_list = []
        for r in range(0, RETRY_MAX):
            try:
                contents_list = self.file_client.list_contents(path)
            except:
                cl, err = sys.exc_info()[:2]
                self.except_handler(err)
            else:
                break
        else:
            # リトライ回数オーバー
            self.except_handler(err)

        for contents in contents_list:

            if stat.S_ISDIR(contents["stat"][stat.ST_MODE]):
                item = self._make_dir_item(path, contents)
                yield item
                for item in self._get_list(path + self.sep + contents["name"]):
                    yield item

            else:
                item = self._make_file_item(path, contents)
                yield item

    def get_list(self, path):

        rel_path = path

        _path = self.__get_server_path(path)

        retlist = []

        for i in self._get_list(_path):
            i.name = i.name[len(rel_path):]
            retlist.append(i)

        return retlist
