# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Module to access the Canon HOME-BOX server with REST API."""


try:
    import json
except:
    import simplejson as json

import os
import time
import stat
import calendar
import unicodedata
import StringIO
import httplib
from urllib import (urlencode, quote)
from urllib2 import (
        HTTPError, HTTPPasswordMgrWithDefaultRealm, HTTPDigestAuthHandler)
from ionas.restapi.accessinfo import (AccessInfo, RequestInfo, NeedToRetry)
from pkgaddon.chbox_config import (HTTP_USER_AGENT_TMPL,)


PUT_METHOD = "PUT"
DELETE_METHOD = "DELETE"

DEFAULT_REPOSITORY_ID = "0"
DIRECTORY_PATH_URL_TMPL = "/repository/%s/folder/path"
BASE_DIRECTORY_URL_TMPL = DIRECTORY_PATH_URL_TMPL + "%s"
LIST_CONTENTS_URL_TMPL = BASE_DIRECTORY_URL_TMPL + "/:items"
RENAME_DIRECTORY_URL_TMPL = BASE_DIRECTORY_URL_TMPL + "/:name"
MOVE_DIRECTORY_URL_TMPL = BASE_DIRECTORY_URL_TMPL + "/:move"
FILE_PATH_URL_TMPL = "/repository/%s/file/path"
BASE_FILE_URL_TMPL = FILE_PATH_URL_TMPL + "%s"
RENAME_FILE_URL_TMPL = BASE_FILE_URL_TMPL + "/:name"
MOVE_FILE_URL_TMPL = BASE_FILE_URL_TMPL + "/:move"
READ_FILE_URL_TMPL = BASE_FILE_URL_TMPL + "/:entity"

BASE_MODE = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
FILE_MODE = BASE_MODE | stat.S_IFREG
DIR_MODE = (BASE_MODE | stat.S_IFDIR |
                stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
MTIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

LINE_END = "\r\n"
BOUNDARY_PREFIX = "--"
BOUNDARY_MAIN = "HOME_BOX_BOUDARY"
BOUNDARY_SUFFIX = "--"

MAX_RETRY_COUNT = 3

UNICODE_LOWER_MISMATCH_LIST = (unichr(0x1E9E),)


class CHBoxHTTPDigestAuthHandler(HTTPDigestAuthHandler):
    """Class for HTTP Digest Authorization of the Canon HOME-BOX."""

    def http_error_401(self, req, fp, code, msg, headers):
        """Try HTTP Digest Authorization for the Canon HOME-BOX."""

        try:
            data = req.get_data()
            if hasattr(data, "reset_stream"):
                data.reset_stream(include_data=True)
                req.add_header("Content-length", len(data))

            return HTTPDigestAuthHandler.http_error_401(
                                        self, req, fp, code, msg, headers)
        except HTTPError, err:
            if err.code != httplib.UNAUTHORIZED:
                self.reset_retry_count()
            raise


class DataStream:
    """Class for accessing data as stream."""

    def __get_length(self, data, length):
        """Return a length of "data"."""

        if length > 0:
            return length

        try:
            return len(data)
        except:
            pass

        try:
            length = 0
            if hasattr(data, "seek") and hasattr(data, "tell"):
                data.seek(0, os.SEEK_END)
                length = data.tell()
                data.seek(0)

            return length
        except:
            pass

        return 0

    def __get_data_stream(self, data):
        """Return the stream instance to access to "data"."""

        if hasattr(data, "read"):
            return data

        data_stream = StringIO.StringIO()
        data_stream.write(data)
        data_stream.seek(0)
        return data_stream

    def __init__(self, data, length=0):
        """Constructor."""

        self.data_stream = self.__get_data_stream(data)
        self.length = self.__get_length(data, length)

    def __len__(self):
        """Return a length of the stream."""

        return self.length

    def read(self, size=-1):
        """Return the data."""

        return self.data_stream.read(size)

    def seek(self, offset, whence=os.SEEK_SET):
        return self.data_stream.seek(offset, whence)


class UploadFileData:
    """Class for upload file data by "multipart/form-data"."""

    def __get_form_data_string(self, name, data=None, type=None):
        """Return the prefix string data for uploading."""

        form_data = BOUNDARY_PREFIX + self.boundary + LINE_END
        form_data += 'Content-Disposition: form-data; name="%s"%s' % \
                        (name, LINE_END)
        if type != None:
            form_data += "Content-Type: %s%s" % (type, LINE_END)
        if data != None:
            form_data += LINE_END + data + LINE_END
        return form_data

    def __get_prefix_string(self, path, overwrite):
        """Return the prefix string data for uploading."""

        prefix = ""
        if not overwrite:
            prefix += self.__get_form_data_string(
                                "parentFolderPath", os.path.dirname(path))
            prefix += self.__get_form_data_string(
                                "name", os.path.basename(path))

        prefix += self.__get_form_data_string(
                                "entity", type="application/octet-stream")
        prefix += LINE_END
        return prefix

    def __get_suffix_string(self):
        """Return the suffix string data for uploading."""

        return LINE_END + BOUNDARY_PREFIX + self.boundary + \
                        BOUNDARY_SUFFIX + LINE_END

    def __read(self, stream, size=-1):
        """Return the data in "stream"."""

        data = ""
        read_size = size
        while True:
            buf = stream.read(read_size)
            if len(buf) <= 0:
                return data
            data += buf
            if size > 0:
                read_size = size - len(buf)

    def __init__(self, path, data, length, boundary, overwrite=False):
        """Constructor."""

        self.boundary = boundary
        self.headers = {"Content-Type":
                            "multipart/form-data; boundary=%s" % (boundary,)}
        self.prefix = DataStream(
                                self.__get_prefix_string(path, overwrite))
        self.main = DataStream(data, length)
        self.suffix = DataStream(self.__get_suffix_string())

        self.length = 0
        self.stream_list = ()
        self.reset_stream()

    def __len__(self):
        """Return a length of the upload data."""

        return self.length

    def get_headers(self):
        """Return dictionary for the request header."""

        return self.headers

    def reset_stream(self, include_data=False):
        """Reset the upload data stream."""

        if include_data:
            self.stream_list = (self.prefix, self.main, self.suffix)
            self.length = len(self.prefix) + len(self.main) + len(self.suffix)
        else:
            self.stream_list = (self.prefix, self.suffix)
            self.length = len(self.prefix) + len(self.suffix)

        self.prefix.seek(0)
        self.main.seek(0)
        self.suffix.seek(0)
        self.headers["Content-length"] = self.length

    def read(self, size=-1):
        """Return the upload data."""

        data = ""
        read_size = size
        for stream in self.stream_list:
            data += self.__read(stream, read_size)
            if size >= 0 and len(data) == size:
                return data

            if size > 0:
                read_size = size - len(data)

        return data


class CHBoxInfo(AccessInfo):
    """Class to access to the Canon HOME-BOX server."""

    def __get_headers(self):
        """Return headers used in common requests."""

        return {
            "User-Agent": HTTP_USER_AGENT_TMPL,
            "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
        }

    def __to_str(self, input):
        """Return string accoding to input."""

        if hasattr(input, "read"):
            return input.read()

        return input

    def __to_utf8(self, string):
        """Convert "string" to UTF-8 string if it is "Unicode"."""

        if isinstance(string, unicode):
            return string.encode("UTF-8")
        return string

    def __to_unicode(self, string):
        """Convert "string" to Unicode string if it is "str".
           Note: character code must be "UTF-8" if "string" is unicode.
        """

        if isinstance(string, str):
            return string.decode("UTF-8")
        return string

    def __quote(self, string):
        """Convert "string" to UTF-8 string if it is "Unicode" and quote it."""

        return quote(self.__to_utf8(string))

    def __get_error_code(self, response):
        """Return error code in response."""

        try:
            err_info = json.loads(response)
            return err_info["errorCode"]
        except:
            pass

        return ""

    def __init__(self, base_url, portal, user, password, application_id,
            repository_id=DEFAULT_REPOSITORY_ID):
        """Constructor."""

        self.base_url = base_url
        self.portal = portal
        self.user = user
        self.password = password
        self.application_id = application_id
        self.repository_id = repository_id

        passwd_mgr = HTTPPasswordMgrWithDefaultRealm()
        passwd_mgr.add_password(
            None, self.base_url, self.user + "@" + self.portal, self.password)
        self.url_handlers = [CHBoxHTTPDigestAuthHandler(passwd_mgr)]

    def get_url_handlers(self):
        """Return a handler list to open URL for request.."""

        return self.url_handlers

    def to_content_list(self, response):
        """Convert "response" to the content list."""

        contents = json.loads(self.__to_str(response))
        content_list = []
        for content in contents["items"]:
            content_info = {"name": content["name"],
                            "stat": self.to_stat_info(json.dumps(content))}

            content_list.append(content_info)

        return content_list

    def convert_path(self, path):
        """Convert charcter in "path" according to the server rules."""

        is_decoded = False
        if isinstance(path, str):
            try:
                is_decoded = True
                path = self.__to_unicode(path)
            except:
                # Do nothing
                return path

        # Normalize "path" according to unicode "NFC" rules."""
        normalized_path = unicodedata.normalize("NFC", path)
        if is_decoded:
            normalized_path = normalized_path.encode("UTF-8")
        return normalized_path

    def compare_path(self, path, path_on_server):
        """Compare "path" and "path_on_server".
           "path" is converted according to the server rules.
        """

        upper_path = self.convert_path(self.__to_unicode(path)).upper()
        upper_path_on_server = self.__to_unicode(path_on_server).upper()

        if len(upper_path) != len(upper_path_on_server):
            return False

        for idx in range(0, len(upper_path)):
            if upper_path[idx] == upper_path_on_server[idx]:
                continue

            if upper_path[idx] in UNICODE_LOWER_MISMATCH_LIST:
                return False

            if upper_path_on_server[idx] in UNICODE_LOWER_MISMATCH_LIST:
                return False

            if upper_path[idx].lower() == upper_path_on_server[idx].lower():
                continue

            return False

        return True

    def to_stat_info(self, response):
        """Convert "response" to the stat information."""

        json_stat_info = json.loads(self.__to_str(response))
        if json_stat_info["itemType"] == "folder":
            mode = DIR_MODE
            size = 0
        else:
            mode = FILE_MODE
            size = json_stat_info["fileSize"]

        utc_time_tuple = time.strptime(json_stat_info["modifiedTime"],
                                        MTIME_FORMAT)
        mtime = int(calendar.timegm(utc_time_tuple))

        return {
                stat.ST_MODE: mode,
                stat.ST_MTIME: mtime,
                stat.ST_SIZE: size,
            }

    def handle_exception(self, exc, request_info, retry_count):
        """Handle "exc" exception."""

        try:
            raise exc
        except HTTPError, err:
            if (err.code == httplib.GATEWAY_TIMEOUT and
                    retry_count < MAX_RETRY_COUNT):
                raise NeedToRetry(request_info)

            raise

    def get_list_contents_req_info(self, path):
        """Return the "RequestInfo" to list contents in "path" directory."""

        url = self.base_url + LIST_CONTENTS_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        return RequestInfo(url, headers=self.__get_headers())

    def get_make_directory_req_info(self, path):
        """Return the "RequestInfo" to make "path" directory."""

        url = self.base_url + DIRECTORY_PATH_URL_TMPL % (self.repository_id,)
        param = {"parentFolderPath": self.__to_utf8(os.path.dirname(path)),
                        "name": self.__to_utf8(os.path.basename(path))}
        return RequestInfo(url, urlencode(param), headers=self.__get_headers())

    def get_delete_directory_req_info(self, path):
        """Return the "RequestInfo" to delete "path" directory."""

        url = self.base_url + BASE_DIRECTORY_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        return RequestInfo(url, headers=self.__get_headers(),
                            method=DELETE_METHOD)

    def get_stat_directory_req_info(self, path):
        """Return the "RequestInfo" to get status of "path" directory."""

        url = self.base_url + BASE_DIRECTORY_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        return RequestInfo(url, headers=self.__get_headers())

    def get_rename_directory_req_info(self, path, new_name):
        """Return the "RequestInfo" to rename "path" directory."""

        url = self.base_url + RENAME_DIRECTORY_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        param = {"name": self.__to_utf8(new_name)}
        return RequestInfo(
            url,
            urlencode(param),
            headers=self.__get_headers(),
            method=PUT_METHOD,
        )

    def get_move_directory_req_info(self, path, parent_path):
        """Return the "RequestInfo" to move "path" directory."""

        url = self.base_url + MOVE_DIRECTORY_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        param = {"parentFolderPath": self.__to_utf8(parent_path)}
        return RequestInfo(url, urlencode(param), headers=self.__get_headers())

    def get_delete_file_req_info(self, path):
        """Return the "RequestInfo" to delete "path" file."""

        url = self.base_url + BASE_FILE_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        return RequestInfo(url, headers=self.__get_headers(),
                            method=DELETE_METHOD)

    def get_stat_file_req_info(self, path):
        """Return the "RequestInfo" to get status of "path" file."""

        url = self.base_url + BASE_FILE_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        return RequestInfo(url, headers=self.__get_headers())

    def get_rename_file_req_info(self, path, new_name):
        """Return the "RequestInfo" to rename "path" file."""

        url = self.base_url + RENAME_FILE_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        param = {"name": self.__to_utf8(new_name)}
        return RequestInfo(
            url,
            urlencode(param),
            headers=self.__get_headers(),
            method=PUT_METHOD,
        )

    def get_move_file_req_info(self, path, parent_path):
        """Return the "RequestInfo" to move "path" file."""

        url = self.base_url + MOVE_FILE_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        param = {"parentFolderPath": self.__to_utf8(parent_path)}
        return RequestInfo(url, urlencode(param), headers=self.__get_headers())

    def get_read_file_req_info(self, path):
        """Return the "RequestInfo" to read from "path" file."""

        url = self.base_url + READ_FILE_URL_TMPL % \
                    (self.repository_id, self.__quote(path))
        return RequestInfo(url, headers=self.__get_headers())

    def get_write_file_req_info(self, path, data, length=0, overwrite=False):
        """Return the "RequestInfo" to write to "path" file."""

        url = self.base_url + FILE_PATH_URL_TMPL % (self.repository_id,)
        if overwrite:
            url = self.base_url + READ_FILE_URL_TMPL % \
                        (self.repository_id, self.__quote(path))

        upload_file_data = UploadFileData(self.__to_utf8(path), data, length,
                                BOUNDARY_MAIN, overwrite)

        req_headers = self.__get_headers()
        req_headers.update(upload_file_data.get_headers())
        return RequestInfo(url, upload_file_data,
                        headers=req_headers)
