#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Copyright (C) I-O DATA DEVICE, INC.  All rights reserved.
"""
from __future__ import with_statement

import os
import re
import logging
import socket
import errno
import pwd
import grp
import time
import shutil
import urllib2
import locale
import json
import datetime as dt
from sys import exc_info
from StringIO import StringIO
import traceback as tb
from product_conf import (CONF_DIR, TEMP_DIR, LOG_DIR,
                          STOR_IF_LIST, STOR_IF_TO_VOL)
from product_conf.devinfo import VOL_INFO
from ionas.nascmd import NasCmd
from chbox.chboxfiletree  import ChboxFileTree as ServerFileTree
from chbox.chboxlocalfiletree  import (ChboxLocalFileTree, EQUIVALENT_SKIP_LOG)
from ionas.fileaccess import DEFAULT_LOCALE
from ionas.fileaccess import LocalAccess
from fal_info import (ReqTypes, Results, FAL_SHARES_PATH, FAL_CHBOX_PATH)
from fal_msginfo import Msg, Cat
from chbox.chboxutil import (clear_announce,
                             save_event_log,
                             get_proxy,
                             send_request,
                             get_tmp_path,
                             is_share_available,
                             get_sync_log_dir_path)
from chbox.chboxaccess import (CHBOX_SEP)
from ionas.filesync.sync import (EconomySyncInfo, EconomySyncEx,
                                 SyncMgr, debug,
                                 UPLOAD_TIME_INFO, DOWNLOAD_TIME_INFO)
from ionas.filesync import (SKIP_ERROR_LOG_LV, FILE_SIZE_OVER_LOG)

from ionas.exception import (BaseAccessException,
                             NoEntryError, NoRootEntryError,
                             InvalidFileSystemError, NoSpaceError,
                             ExistsError, SizeTooLargeError,
                             InternalError, NameTooLongError,
                             PermissionError)

from chbox.chboxfiletree import (OperationLocked, ForbiddenNameError)
from ionas.filesync.basefilesync import (IgnoreTargetFile)
from iobase.exception import (NotFoundError, IoBaseException)
from ionas.filesync import (
    TYPE_FILE_TRANSFER_OK, TYPE_FILE_TRANSFER_NG, TYPE_SYNC_SKIP,
    OK_LOG, NG_LOG, FILE_DELETE_LOG, DIR_CREATE_LOG, DIR_DELETE_LOG)
from ionas.filesync import (MODE_CH_FILE, MODE_CH_DIR)
import chbox_summary_proto as sum_proto
from pkgaddon.chbox_config import REST_BASE_HOST
import ionas.exception
import ionas.filesync.basefilesync

ionas.filesync.basefilesync.EXCUSE_TIME = 30

SERVICE_NAME = "chbox"

NAME_NAME = "name"
PATH_NAME = "path"
SHARESERVICE_NAME = "shareservice"
CHBOXSYNC_PATH = SHARESERVICE_NAME + "/chboxsync"

PORTALID_NAME = "portalid"
LOGINID_NAME = "loginid"
LOINPASSWD_NAME = "passwd"
ROOT_PATH_XPATH = "folder/path"
UPLOAD_MAX_SIZE_NAME = "uploadmaxfilesize"

HTTPMETHOD_NAME = "httpmethod"
STORAGECLASS_NAME = "storageclass"
DOWNLOADINTERVAL_NAME = "downloadinterval"
ENDPOINT_NAME = "endpoint"

SYNC_INFO_DIR = os.path.join(CONF_DIR, SERVICE_NAME, "info")

UID_NAME = "uid"
GID_NAME = "gid"

STATUS_GATEWAY_TIMEOUT = 504

ACCESSKEY_INFO = "accesskey"
SECRETKEY_INFO = "secretkey"
PROXY_HOST_INFO = "proxy_host"
PROXY_PORT_INFO = "proxy_port"
PORT_INFO = "port"
IS_SECURE_INFO = "is_secure"
BUCKET_INFO = "bucket"
ENDPOINT_INFO = "endpoint"

DRAWER_INFO = "drawer"
ENDPOINT_INFO = "endpoint"
UPLOAD_TIME_KEY = "upload_time"
DOWNLOAD_TIME_KEY = "download_time"

TARGET_ROOT_INFO = "target_root"
PORTAL_INFO = "portal"

XEROXWF_WEBAPI_HOST = "workingfolder-api.fujixerox.com"

INTERNAL_NAME = STOR_IF_TO_VOL[STOR_IF_LIST[0]][1]
UPLOADED_PATH = os.path.join(VOL_INFO[INTERNAL_NAME]["mnt_point"],
                             "chbox_uploaded")

IMMEDIATE_DOWNLOAD_INFO_DIR = \
    os.path.join(TEMP_DIR, "chbox", "immediate_download")

UPLOADING = "uploading"
DOWNLOADING = "downloading"

SHARE_CHECK_PERIOD = 60
RETRY_PERIOD = 600
RECOVER_PERIOD = 10
RECOVER_CNT = 3

# LOG TYPE
NO_ROOT_ENTRY_ERR_LOG = "NO_ROOT_ENTRY_ERR_LOG"
NO_SPACE_ON_LOCAL_ERR_LOG = "NO_SPACE_ON_LOCAL_ERR_LOG"
NO_SPACE_ON_SERVER_ERR_LOG = "NO_SPACE_ON_SERVER_ERR_LOG"
INVALID_IDENTIFIER_ERR_LOG = "INVALID_IDENTIFIER_ERR_LOG"
INVALID_PASSWORD_ERR_LOG = "INVALID_PASSWORD_ERR_LOG"
INVALID_FS_ERR_LOG = "INVALID_FS_ERR_LOG"
FILE_TOO_LARGE_ERR_LOG = "FILE_TOO_LARGE_ERR_LOG"
SKIP_ERR_LOG = "SKIP_ERR_LOG"
TIMEOUT_ERR_LOG = "TIMEOUT_ERR_LOG"
MISC_ERR_LOG = "MISC_ERR_LOG"
SESSION_QUOTA_EXCEEDED_ERR_LOG = "SESSION_QUOTA_EXCEEDED_ERR_LOG"
FORBIDDEN_FILE_NAME_ERR_LOG = "FORBIDDEN_FILE_NAME_ERR_LOG"
NO_PERMISSION_ERR_LOG = "NO_PERMISSION_ERR_LOG"
LOCKED_ERR_LOG = "ERR_LOCKED_ERR_LOG"
CONNECT_ERR_LOG = "CONNECT_ERR_LOG"
SAME_FILE_NAME_EXISTS_ERR_LOG = "SAME_FILE_NAME_EXISTS_ERR_LOG"
TOO_LONG_FILE_NAME_ERR_LOG = "TOO_LONG_FILE_NAME_ERR_LOG"

ERR_LOG_MAP = \
    {NO_ROOT_ENTRY_ERR_LOG: Msg.ERR_CHBOX_FOLDER_NOT_FOUND,
     NO_SPACE_ON_LOCAL_ERR_LOG: Msg.ERR_CHBOX_NO_SPACE_ON_LOCAL,
     NO_SPACE_ON_SERVER_ERR_LOG: Msg.ERR_CHBOX_NO_SPACE_ON_SERVER,
     INVALID_FS_ERR_LOG: Msg.ERR_CHBOX_INVALID_FORMAT,
     FILE_TOO_LARGE_ERR_LOG: Msg.ERR_CHBOX_TOO_LARGE_FILE,
     SKIP_ERR_LOG: Msg.ERR_CHBOX_SYNC_SKIP,
     TIMEOUT_ERR_LOG: Msg.ERR_CHBOX_TIMEOUT,
     MISC_ERR_LOG: Msg.ERR_CHBOX_SYNC,
     SESSION_QUOTA_EXCEEDED_ERR_LOG: Msg.ERR_CHBOX_SESSION_QUOTA_EXCEEDED,
     FORBIDDEN_FILE_NAME_ERR_LOG: Msg.ERR_CHBOX_FORBIDDEN_FILE_NAME,
     NO_PERMISSION_ERR_LOG: Msg.ERR_CHBOX_NO_PERMISSION,
     LOCKED_ERR_LOG: Msg.ERR_CHBOX_LOCKED,
     CONNECT_ERR_LOG: Msg.ERR_CHBOX_CONNECT,
     SAME_FILE_NAME_EXISTS_ERR_LOG: Msg.ERR_CHBOX_SAME_FILE_NAME_EXISTS,
     TOO_LONG_FILE_NAME_ERR_LOG: Msg.ERR_CHBOX_TOO_LONG_FILE_NAME,
     }

ERR_MARK = "+++ Error +++"
ERR_TEXT_FMT = "%s " + ERR_MARK + "\r\n%s"
LOG_FMT_STR = "%s -> %s\r\n"

SUCCESS = "success"
FAILED = "failed"
STOPPED = "stopped"

WRITE_MODE = "write"
READ_MODE = "read"
APPEND_MODE = "append"

SYNC_DIR_FMT_STR = "%04d_%02d_%02d-%02d_%02d_%02d"
MAX_FAIL_FILE_COUNT = 1000
LIMIT_MEM_LOG_LENGTH = 10 * 1024 * 1024  # 10Mbyte

LOG_DIR_NAME = "HOME-BOX"
SYNC_LOG_DIR = os.path.join(get_sync_log_dir_path(), LOG_DIR_NAME)

UPLOAD_COUNT = "upload"
DOWNLOAD_COUNT = "download"

LOG_SUFFIX = ".log"
LOG_RE = "^[0-9\_\-]{19}\.log$"
SUMMAEY_RE = "^[0-9\_\-]{19}\_summary.log$"
FAIL_LOG_SUFFIX = "_failed.log"
STOP_LOG_SUFFIX = "_stopped.log"
SUMMARY_LOG_SUFFIX = "_summary.log"
MAX_LOG_COUNT = 20

LOG_TITLE = \
"****************************************\r\n" + \
"   Sync with Canon HOME-BOX Log\r\n" + \
"****************************************\r\n"

LOG_LOCAL_SHARE_TMPL = "Local Share    : '%s'\r\n"
LOG_SERVER_FOLDER_TMPL = "Server Folder  : '%s'\r\n"
LOG_START_DAYTIME_TMPL = "Start Day Time : %s\r\n"

LOG_START_UPLOAD = \
    "\r\n///// Start uploading files from local ==> cloud /////\r\n"

LOG_END_UPLOAD = \
    "///// End uploading /////\r\n"

LOG_START_DOWNLOAD = \
    "\r\n///// Start downloading files from cloud ==> local /////\r\n"

LOG_END_DOWNLOAD = \
    "///// End downloading /////\r\n"

LOG_CANCEL_SYNC = "///// Cancel Sync /////\r\n"

LOG_EXACT_FIRST_UPLOAD = "///// Exact First Upload /////\r\n"
LOG_FIRST_UPLOAD = "///// First Upload /////\r\n"
LOG_NORMAL_UPLOAD = "///// Normal Upload /////\r\n"
LOG_ABORT_UPLOAD = "///// Abort Upload ///// \r\n"

LOG_EXACT_FIRST_DOWNLOAD = "///// Exact First Download ///// \r\n"
LOG_FIRST_DOWNLOAD = "///// First Download ///// \r\n"
LOG_NORMAL_DOWNLOAD = "///// Normal Download ///// \r\n"
LOG_ABORT_DOWNLOAD = "///// Abort Download ///// \r\n"

RESULT_DIR_NAME = "result"

RESULT_DIR_NAME = "result"


class ServerCommunicationError(BaseAccessException):
    pass


JSON_ERR_ID_KEY = u"errorCode"

errid_except_dict = {
    "CONFLICT": ExistsError,
    "NO_AUTHORIZATION": ServerCommunicationError,
    "UNAUTHORIZATION": ServerCommunicationError,
    "FORBIDDEN": PermissionError,
    "NOT_FOUND": NoEntryError,
    "LOCKED": OperationLocked,
    "PORTAL_INSUFFICIENT_STORAGE": NoSpaceError,
    "SYSTEM_INSUFFICIENT_STORAGE": NoSpaceError,
    "UNAUTHORIZED_SESSION": ServerCommunicationError,
    "BAD_REQUEST_PARAMETER": ForbiddenNameError,
    "BAD_REQUEST_URI": InternalError,
    "BAD_SESSION": InternalError,
    "METHOD_NOT_ALLOWED": InternalError,
    "NOT_ACCEPTABLE": InternalError,
    "UNKNOWN_ERROR": InternalError,
    "ERROR": InternalError,
    "ERROR_ROLLBACK_FAILED": InternalError,
    "PORTAL_INVALID_USER_MAX_COUNT": InternalError,
    "GONE": InternalError,
    }

# HTTP status
UNAUTHORIZED_HTTP_STATUS = 401

def lf2crlf(lf_str):

    return lf_str.replace("\r\n", "\n").replace("\n", "\r\n")


class CreateLogFileError(IoBaseException):
    pass


class Summary():

    def __init__(self, share_name):
        self.__share_name = share_name
        self.__share_map = {}
        self.__ng_files_map = {}
        self.__ng_dirs_map = {}
        self.__ng_del_files_map = {}
        self.__ng_del_dirs_map = {}

        self.__share_map[self.__share_name] = \
            {UPLOAD_COUNT: {sum_proto.SUMMARY_OK_KEY: 0,
                            sum_proto.SUMMARY_NG_KEY: 0,
                            sum_proto.SUMMARY_FILE_DEL_OK_KEY: 0,
                            sum_proto.SUMMARY_FILE_DEL_NG_KEY: 0,
                            sum_proto.SUMMARY_DIR_OK_KEY: 0,
                            sum_proto.SUMMARY_DIR_NG_KEY: 0,
                            sum_proto.SUMMARY_DIR_DEL_OK_KEY: 0,
                            sum_proto.SUMMARY_DIR_DEL_NG_KEY: 0,
                            },
             DOWNLOAD_COUNT: {sum_proto.SUMMARY_OK_KEY: 0,
                              sum_proto.SUMMARY_NG_KEY: 0,
                              sum_proto.SUMMARY_FILE_DEL_OK_KEY: 0,
                              sum_proto.SUMMARY_FILE_DEL_NG_KEY: 0,
                              sum_proto.SUMMARY_DIR_OK_KEY: 0,
                              sum_proto.SUMMARY_DIR_NG_KEY: 0,
                              sum_proto.SUMMARY_DIR_DEL_OK_KEY: 0,
                              sum_proto.SUMMARY_DIR_DEL_NG_KEY: 0, }}

        self.__ng_files_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                  DOWNLOAD_COUNT: []}
        self.__ng_dirs_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                 DOWNLOAD_COUNT: []}
        self.__ng_del_files_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                      DOWNLOAD_COUNT: []}
        self.__ng_del_dirs_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                     DOWNLOAD_COUNT: []}

        try:
            self.default_encode = locale.getdefaultlocale()[1]
            if not self.default_encode:
                self.default_encode = DEFAULT_LOCALE
        except:
            self.default_encode = DEFAULT_LOCALE

    def init_summay_param(self):

        self.__share_map = {}
        self.__ng_files_map = {}
        self.__share_map[self.__share_name] = \
            {UPLOAD_COUNT: {sum_proto.SUMMARY_OK_KEY: 0,
                            sum_proto.SUMMARY_NG_KEY: 0,
                            sum_proto.SUMMARY_FILE_DEL_OK_KEY: 0,
                            sum_proto.SUMMARY_FILE_DEL_NG_KEY: 0,
                            sum_proto.SUMMARY_DIR_OK_KEY: 0,
                            sum_proto.SUMMARY_DIR_NG_KEY: 0,
                            sum_proto.SUMMARY_DIR_DEL_OK_KEY: 0,
                            sum_proto.SUMMARY_DIR_DEL_NG_KEY: 0,
                            },
             DOWNLOAD_COUNT: {sum_proto.SUMMARY_OK_KEY: 0,
                              sum_proto.SUMMARY_NG_KEY: 0,
                              sum_proto.SUMMARY_FILE_DEL_OK_KEY: 0,
                              sum_proto.SUMMARY_FILE_DEL_NG_KEY: 0,
                              sum_proto.SUMMARY_DIR_OK_KEY: 0,
                              sum_proto.SUMMARY_DIR_NG_KEY: 0,
                              sum_proto.SUMMARY_DIR_DEL_OK_KEY: 0,
                              sum_proto.SUMMARY_DIR_DEL_NG_KEY: 0, }}

        self.__ng_files_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                  DOWNLOAD_COUNT: []}
        self.__ng_dirs_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                 DOWNLOAD_COUNT: []}
        self.__ng_del_files_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                      DOWNLOAD_COUNT: []}
        self.__ng_del_dirs_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                     DOWNLOAD_COUNT: []}

    def set_sync_start_time(self, start_daytime):
        self.__start_daytime = start_daytime

    def set_sync_end_time(self, end_daytime):
        self.__end_daytime = end_daytime

    def __get_count_target_key(self, log_info):

        try:
            # ファイル作成
            if log_info.type == TYPE_FILE_TRANSFER_OK:
                return sum_proto.SUMMARY_OK_KEY

            if log_info.type == TYPE_FILE_TRANSFER_NG or \
               log_info.exc_type == SizeTooLargeError or \
               log_info.exc_type == IgnoreTargetFile:
                return sum_proto.SUMMARY_NG_KEY

            # ファイル削除
            if log_info.log.startswith(FILE_DELETE_LOG + " " + OK_LOG):
                return sum_proto.SUMMARY_FILE_DEL_OK_KEY

            if log_info.log.startswith(FILE_DELETE_LOG + " " + NG_LOG):
                return sum_proto.SUMMARY_FILE_DEL_NG_KEY

            # ディレクトリ作成
            if log_info.log.startswith(DIR_CREATE_LOG + " " + OK_LOG):
                return sum_proto.SUMMARY_DIR_OK_KEY

            if log_info.log.startswith(DIR_CREATE_LOG + " " + NG_LOG):
                return sum_proto.SUMMARY_DIR_NG_KEY

            # ディレクトリ削除
            if log_info.log.startswith(DIR_DELETE_LOG + " " + OK_LOG):
                return sum_proto.SUMMARY_DIR_DEL_OK_KEY

            if log_info.log.startswith(DIR_DELETE_LOG + " " + NG_LOG):
                return sum_proto.SUMMARY_DIR_DEL_NG_KEY
        except:
            pass

        return None

    def __get_ng_list(self, tgt_key, share_name, direction):

        if tgt_key == sum_proto.SUMMARY_NG_KEY:
            return self.__ng_files_map[share_name][direction]

        elif tgt_key == sum_proto.SUMMARY_FILE_DEL_NG_KEY:
            return self.__ng_del_files_map[share_name][direction]

        elif tgt_key == sum_proto.SUMMARY_DIR_NG_KEY:
            return self.__ng_dirs_map[share_name][direction]

        elif tgt_key == sum_proto.SUMMARY_DIR_DEL_NG_KEY:
            return self.__ng_del_dirs_map[share_name][direction]

        return None

    def __get_ng_list_from_ok_key(self, tgt_key, share_name, direction):

        if tgt_key == sum_proto.SUMMARY_OK_KEY:
            return self.__ng_files_map[share_name][direction]

        elif tgt_key == sum_proto.SUMMARY_FILE_DEL_OK_KEY:
            return self.__ng_del_files_map[share_name][direction]

        elif tgt_key == sum_proto.SUMMARY_DIR_OK_KEY:
            return self.__ng_dirs_map[share_name][direction]

        elif tgt_key == sum_proto.SUMMARY_DIR_DEL_OK_KEY:
            return self.__ng_del_dirs_map[share_name][direction]

        return None

    def update(self, log_info):

        if isinstance(log_info.dst, ServerFileTree):
            direction = UPLOAD_COUNT
        else:
            direction = DOWNLOAD_COUNT

        tgt_key = self.__get_count_target_key(log_info)
        if not tgt_key:
            return

        file_path = log_info.relpath.replace("/", "\\")

        ng_list = self.__get_ng_list(tgt_key, self.__share_name, direction)

        if ng_list is not None:
            if file_path not in ng_list:
                if len(ng_list) < MAX_FAIL_FILE_COUNT:
                    ng_list.append(file_path)

                self.__share_map[self.__share_name][direction][tgt_key] += 1

        else:
            self.__share_map[self.__share_name][direction][tgt_key] += 1

            ng_list = self.__get_ng_list_from_ok_key(
                          tgt_key, self.__share_name, direction)

            if ng_list is None:
                return

            # リライで成功した場合
            if file_path in ng_list:
                ng_list.remove(file_path)

                self.__share_map[self.__share_name][
                    direction][sum_proto.SUMMARY_NG_KEY] -= 1

    def get_summary_body(self, is_download=False):

        share = self.__share_name

        NG_FILES_IDX = 0
        NG_DIRS_IDX = 1
        NG_DEL_FILES_IDX = 2
        NG_DEL_DIRS_IDX = 3

        if is_download:
            count_type = DOWNLOAD_COUNT
            TITLE_LST = [sum_proto.SUMMARY_DOWNLOAD_NG_FILES,
                         sum_proto.SUMMARY_DOWNLOAD_NG_DIRS,
                         sum_proto.SUMMARY_DEL_NG_FILES,
                         sum_proto.SUMMARY_DEL_NG_DIRS]
        else:
            count_type = UPLOAD_COUNT
            TITLE_LST = [sum_proto.SUMMARY_UPLOAD_NG_FILES,
                         sum_proto.SUMMARY_UPLOAD_NG_DIRS,
                         sum_proto.SUMMARY_DEL_NG_FILES,
                         sum_proto.SUMMARY_DEL_NG_DIRS]

        cnt_map = self.__share_map[share][count_type]

        body_map = cnt_map

        total_files = cnt_map[sum_proto.SUMMARY_OK_KEY] + \
                      cnt_map[sum_proto.SUMMARY_NG_KEY] + \
                      cnt_map[sum_proto.SUMMARY_FILE_DEL_OK_KEY] + \
                      cnt_map[sum_proto.SUMMARY_FILE_DEL_NG_KEY]

        total_dirs = cnt_map[sum_proto.SUMMARY_DIR_OK_KEY] + \
                     cnt_map[sum_proto.SUMMARY_DIR_NG_KEY] + \
                     cnt_map[sum_proto.SUMMARY_DIR_DEL_OK_KEY] + \
                     cnt_map[sum_proto.SUMMARY_DIR_DEL_NG_KEY]

        body_map.update({sum_proto.TOTAL_FILES: total_files,
                         sum_proto.TOTAL_DIRS: total_dirs})

        if is_download:
            body_str = sum_proto.SUMMARY_DOWNLOAD_BODY_TMPL % body_map
        else:
            body_str = sum_proto.SUMMARY_UPLOAD_BODY_TMPL % body_map

        if len(self.__ng_files_map[share][count_type]) > 0:
            body_str += TITLE_LST[NG_FILES_IDX]
            body_str += \
                unicode("\r\n".join(self.__ng_files_map[share][count_type]),
                        self.default_encode)
            body_str += u"\r\n\r\n"

        if len(self.__ng_dirs_map[share][count_type]) > 0:
            body_str += TITLE_LST[NG_DIRS_IDX]
            body_str += \
                unicode("\r\n".join(self.__ng_dirs_map[share][count_type]),
                        self.default_encode)
            body_str += u"\r\n\r\n"

        if len(self.__ng_del_files_map[share][count_type]) > 0:
            body_str += TITLE_LST[NG_DEL_FILES_IDX]
            body_str += \
                unicode("\r\n".join(
                            self.__ng_del_files_map[share][count_type]),
                        self.default_encode)
            body_str += u"\r\n\r\n"

        if len(self.__ng_del_dirs_map[share][count_type]) > 0:
            body_str += TITLE_LST[NG_DEL_DIRS_IDX]
            body_str += \
                unicode("\r\n".join(self.__ng_del_dirs_map[share][count_type]),
                        self.default_encode)
            body_str += u"\r\n\r\n"

        body_str = body_str.replace("\r\n", "\n").replace("\n", "\r\n")
        return body_str.encode(self.default_encode)

    def get_summary_header(self, result):

        delta_time = self.__end_daytime - self.__start_daytime
        elapse_seconds = delta_time.days * 24 * 3600 + delta_time.seconds
        elapse_time_str = "%02d:%02d:%02d" % \
                              (elapse_seconds / 3600,
                               (elapse_seconds % 3600) / 60,
                               elapse_seconds % 60)
        start_time = self.__start_daytime.strftime('%Y-%m-%d %H:%M:%S')
        end_time = self.__end_daytime.strftime('%Y-%m-%d %H:%M:%S')

        header = {sum_proto.RESULT: result.decode(self.default_encode),
                  sum_proto.SHARE: self.__share_name.decode(
                                       self.default_encode),
                  sum_proto.START: start_time.decode(self.default_encode),
                  sum_proto.END: end_time.decode(self.default_encode),
                  sum_proto.ELAPSE: elapse_time_str.decode(
                                        self.default_encode)}

        header_str = sum_proto.SUMMARY_HEADER_TMPL % header
        header_str = header_str.replace("\r\n", "\n").replace("\n", "\r\n")
        return header_str.encode(self.default_encode)

    def get_file_access_count(self):

        return \
            sum(self.__share_map[self.__share_name][UPLOAD_COUNT].values()) + \
            sum(self.__share_map[self.__share_name][DOWNLOAD_COUNT].values())

    def get_ng_file_count(self):

        cnt_map = self.__share_map[self.__share_name]
        return cnt_map[UPLOAD_COUNT][sum_proto.SUMMARY_NG_KEY] + \
               cnt_map[DOWNLOAD_COUNT][sum_proto.SUMMARY_NG_KEY]


class Logger():

    def __init__(self,
                 share_name,
                 log_file_path,
                 log_accesser,
                 buf_limit_len=LIMIT_MEM_LOG_LENGTH):

        self.__summary = Summary(share_name)
        self.__result = ""
        self.__share_name = share_name
        self.__is_need_save = False

        def _mkdir(path):
            log_accesser.mkdir(path)
            log_accesser.chmod(path, MODE_CH_DIR)

        self._log_open = lambda path, mode:\
            log_accesser.open(path, mode)
        self._log_mkdir = lambda path:\
            _mkdir(path)
        self._log_mv = lambda orig, dst:\
            log_accesser.move(orig, dst)
        self._log_chmod = lambda path, mode:\
            log_accesser.chmod(path, mode)
        self._log_chmod = lambda path, mode:\
            log_accesser.chmod(path, mode)
        self._log_listdir = lambda path:\
            log_accesser.listdir(path)
        self._log_rmdir = lambda path:\
            log_accesser.rmdir(path)
        self._log_stat = lambda path:\
            log_accesser.stat(path)
        self._log_remove = lambda path:\
            log_accesser.remove(path)

        self._log_io = None
        self._log_file_path = log_file_path + LOG_SUFFIX
        self._summary_log_file_path = log_file_path + SUMMARY_LOG_SUFFIX
        self._buf_limit_len = buf_limit_len
        self.make_log_dir()

    def init_log_param(self, log_file_path):

        self.__is_need_save = False
        self._log_file_path = log_file_path + LOG_SUFFIX
        self._summary_log_file_path = log_file_path + SUMMARY_LOG_SUFFIX
        self.__summary.init_summay_param()

    def get_ng_file_count(self):

        try:
            return self.__summary.get_ng_file_count()
        except:
            return 0

    def __write(self, string):

        if not self._log_io:
            self._log_io = StringIO()

        self._log_io.write(string)

        if self._log_io.len > self._buf_limit_len:
            self._flush_and_renew_log_buf()

    def make_log_dir(self):

        sep = "/"
        dir_path = os.path.dirname(self._log_file_path)
        dir_parts = dir_path.split(sep)

        current_path = sep
        for part in dir_parts:
            current_path += sep + part

            try:
                self._log_stat(current_path)
            except NoEntryError:
                try:
                    self._log_mkdir(current_path)
                except ExistsError:
                    pass

    def __flush(self):

        log_file = \
            self._log_open(self._log_file_path, APPEND_MODE)

        self._log_io.seek(0)
        log_file.write(self._log_io)
        log_file.close()

    def _renew_log_buf(self):

        if self._log_io:
            self._log_io.close()
            del self._log_io

        self._log_io = StringIO()

    def _flush_and_renew_log_buf(self):
        """
           Flush log data into log file.
           And free buf and realloc new buf to avoid huge memory consumption.
        """

        self.__is_need_save = False

        try:
            self.__flush()
        except:
            raise CreateLogFileError

        self._log_chmod(self._log_file_path, MODE_CH_FILE)
        self._renew_log_buf()

    def end_log(self, result):

        log_suffix_map = {SUCCESS: LOG_SUFFIX,
                          FAILED: FAIL_LOG_SUFFIX,
                          STOPPED: STOP_LOG_SUFFIX}

        self.__result = result
        suffix = log_suffix_map[result]

        if not self.__is_need_save:
            self._renew_log_buf()
            return

        self._flush_and_renew_log_buf()

        if result in log_suffix_map.keys():
            real_log_path = \
                re.sub(LOG_SUFFIX + "$", suffix, self._log_file_path)
            self._log_mv(self._log_file_path, real_log_path)

        self._save_summary()

        self._delete_oldest_log()

    def set_share(self, share):
        self.__summary.set_share(share)

    def set_sync_start_time(self, start_daytime):
        self.__summary.set_sync_start_time(start_daytime)

    def set_sync_end_time(self, end_daytime):
        self.__summary.set_sync_end_time(end_daytime)

    def __get_detail_info_str(self, log_info):

        DETAIL_STR_TMPL = \
            "HTTPError detail info:\r\n" + \
            "  code : %(code)s\r\n" + \
            "  msg : %(msg)s\r\n" + \
            "  data : %(data)s\r\n"

        try:
            if hasattr(log_info.exc_arg, "detail"):
                err = log_info.exc_arg.detail
                if isinstance(err, urllib2.HTTPError):

                    http_data = ""
                    if hasattr(err, "read"):
                        http_data = err.read()

                    return DETAIL_STR_TMPL % {"code": str(err.code),
                                              "msg": str(err.msg),
                                              "data": http_data}

            if isinstance(log_info.exc_arg, urllib2.HTTPError):
                err = log_info.exc_arg

                http_data = ""
                if hasattr(err, "read"):
                    http_data = err.read()

                return DETAIL_STR_TMPL % {"code": str(err.code),
                                          "msg": str(err.msg),
                                          "data": http_data}
        except:
            pass

        return  ""

    def write_log(self, log_info):
        result = log_info.log
        if result and log_info.type != TYPE_SYNC_SKIP:
            if log_info.lv >= SKIP_ERROR_LOG_LV:
                result = ERR_TEXT_FMT % (result, log_info.trace)

            src = (self.__share_name + '\\' + \
                      log_info.relpath).replace("/", "\\")

            self.__is_need_save = True
            self.__write(lf2crlf(LOG_FMT_STR % (src, result)))

            if log_info.lv >= SKIP_ERROR_LOG_LV:
                detail_str = self.__get_detail_info_str(log_info)
                if detail_str:
                    self.__write(detail_str + "\r\n")

        self.__summary.update(log_info)

    def write(self, data, is_need_save=False):

        self.__write(data)
        if is_need_save:
            self.__is_need_save = True

    def _save_summary(self):

        try:
            self._log_mkdir(os.path.dirname(self._summary_log_file_path))
        except ExistsError:
            pass

        log_file = self._log_open(self._summary_log_file_path, WRITE_MODE)
        log_file.write(StringIO(
            self.__summary.get_summary_header(self.__result)))
        log_file.write(StringIO(self.__summary.get_summary_body(False)))
        log_file.write(StringIO(self.__summary.get_summary_body(True)))
        log_file.close()
        return

    def _delete_oldest_log(self):

        log_dir = os.path.dirname(self._log_file_path)

        try:
            log_list = [log for log in self._log_listdir(log_dir)
                            if not log.endswith(FAIL_LOG_SUFFIX)]
        except:
            return

        for log in sorted(log_list, reverse=True)[MAX_LOG_COUNT:]:
            try:
                self._log_remove(os.path.join(log_dir, log))
            except:
                pass

    def get_file_access_count(self):

        return self.__summary.get_file_access_count()


class ChboxSync(EconomySyncEx):

    SERVICE_ERR_LIST = []

    def __init__(self, info):

        self.info = info

        self.current_direction = None

        self.is_need_retry_sleep = False

        self.nascmd_logging = NasCmd(Results, logger_name="CHBOX",
            log=LOG_DIR + "/chbox.log", log_level=logging.DEBUG)
        now = dt.datetime.now()
        log_file_path = os.path.join(SYNC_LOG_DIR,
                                     self.info.share_name,
                                     self.__get_sync_start_daytime_str(now))
        self._logger = \
            Logger(self.info.share_name, log_file_path, LocalAccess())

        self._callback = ChboxCallBack(self)

        super(ChboxSync, self).__init__(info)

    def __get_sync_start_daytime_str(self, now):

        return SYNC_DIR_FMT_STR % (
            now.year, now.month, now.day, now.hour, now.minute, now.second)

    def __get_err_id(self, httperr_exc):

        try:
            return json.loads(httperr_exc.read())[JSON_ERR_ID_KEY]
        except:
            return None

    def __is_timeout_exception(self, exc):

        try:
            if isinstance(exc, socket.gaierror):
                timeout_args_list = (
                    socket.EAI_NONAME,
                    socket.EAI_NODATA)

                if exc.args[0] in timeout_args_list:
                    return True

            elif isinstance(exc, socket.timeout):
                return True

            elif isinstance(exc, socket.sslerror):
                if exc.args[0] == "The read operation timed out":
                    return True

            elif isinstance(exc, socket.error):
                timeout_args_list = (
                    errno.ETIMEDOUT,
                    errno.EHOSTUNREACH)

                if exc.args[0] in timeout_args_list:
                    return True

                try:
                    if exc.args[1].endswith("(Gateway Time-out)"):
                        return True
                except:
                    pass

            elif isinstance(exc, IOError):
                if isinstance(exc.args[1], socket.timeout):
                    return True

        except:
            return False
        return False

    def _is_misc_recover_err(self, exc):
        # 不明エラーは基本的に
        # recoverする方針(従来どおり)
        return True

    def __get_ev_log(self, exc):
        if self.__is_timeout_exception(exc):
            debug(["!!! timeout err !!!"])
            return ERR_LOG_MAP[TIMEOUT_ERR_LOG]

        if isinstance(exc, urllib2.HTTPError) or \
           isinstance(exc, urllib2.URLError):
            return ERR_LOG_MAP[CONNECT_ERR_LOG]

        if hasattr(exc, "detail"):
            if isinstance(exc.detail, urllib2.HTTPError) or \
               isinstance(exc.detail, urllib2.URLError):
                return ERR_LOG_MAP[CONNECT_ERR_LOG]

        if self._is_misc_recover_err(exc):
            debug(["!!! misc recover err !!!"])
            return ERR_LOG_MAP[MISC_ERR_LOG]

        return None

    def __log_event(self, event, args={}, err_info=""):

        ROOT_UID = 0
        ROOT_GID = 0

        if not event:
            return

        if err_info:
            self.nascmd_logging.error(str(err_info))

        try:
            save_event_log(event, args)
        except:
            try:
                pre_uid = os.geteuid()
                pre_gid = os.getegid()
                os.setegid(ROOT_GID)
                os.seteuid(ROOT_UID)
                save_event_log(event, args)
            finally:
                os.setegid(pre_gid)
                os.seteuid(pre_uid)

    def _init_check(self):
        is_share_available(self.info.share_name, self.info.share_abs_path)

    def _set_trees(self):

        self._local_tree = \
            ChboxLocalFileTree(self.info.share_abs_path,
                               tmp_path=get_tmp_path(self.info.share_abs_path),
                               log_callback=self._callback.log_callback)
        self._server_tree = \
            ServerFileTree(portalid=self.info.portalid,
                           login_id=self.info.login_id,
                           login_passwd=self.info.login_passwd,
                           root=self.info.root,
                           uploadmaxfilesize=self.info.uploadmaxfilesize,
                           proxy_host=self.info.proxy_host,
                           proxy_port=self.info.proxy_port,
                           misc_callback=self._callback)

    def __recover_err_handle(self, exc, do_recover):
        share_name = self.info.share_name

        ev_log = self.__get_ev_log(exc)

        if not ev_log:
            return False

        if do_recover:
            debug(["!!! do recover !!!"])
            time.sleep(RECOVER_PERIOD)
            return True

        self.__log_event(ev_log, {"name": share_name})

        self.is_need_retry_sleep = True

        raise exc, None, exc_info()[2]

    def __no_retry_err_handle(self, exc):
        share_name = self.info.share_name

        if isinstance(exc, InvalidFileSystemError):
            self.__log_event(ERR_LOG_MAP[INVALID_FS_ERR_LOG],
                            {"name": share_name})
        else:
            return

        self._set_disable_share(share_name)

        debug(["!!! no retry err !!!"])

        raise exc, None, exc_info()[2]

    def __retry_err_handle(self, exc):
        share_name = self.info.share_name
        if isinstance(exc, NoSpaceError) and \
                 self.current_direction == DOWNLOADING:
            self.__log_event(ERR_LOG_MAP[NO_SPACE_ON_LOCAL_ERR_LOG],
                     {"name": share_name})

        elif isinstance(exc, NoSpaceError) and \
               isinstance(exc.detail, urllib2.HTTPError):
            self.__log_event(ERR_LOG_MAP[NO_SPACE_ON_SERVER_ERR_LOG],
                     {"name": self.info.portalid + ":" + self.info.root})

        elif isinstance(exc, ServerCommunicationError):
            self.__log_event(ERR_LOG_MAP[SESSION_QUOTA_EXCEEDED_ERR_LOG],
                     {"name": share_name})

        elif isinstance(exc, NoRootEntryError):
            self.__log_event(ERR_LOG_MAP[NO_ROOT_ENTRY_ERR_LOG],
                    {"name": self.info.portalid + ":" + self.info.root})
        else:
            return

        debug(["!!! retry err !!!"])
        self.is_need_retry_sleep = True
        raise exc, None, exc_info()[2]

    def __misc_err_handle(self, exc):
        self.__log_event(ERR_LOG_MAP[MISC_ERR_LOG],
                        {"name": self.info.share_name},
                        tb.format_exc())
        raise exc, None, exc_info()[2]

    def _init_except_handler(self, exc):

        self.__no_retry_err_handle(exc)
        self.__misc_err_handle(exc)

    def _access_except_handler(self, exc, is_do_recover=False):

        self.__no_retry_err_handle(exc)
        self.__retry_err_handle(exc)
        if self.__recover_err_handle(exc, is_do_recover):
            return True
        self.__misc_err_handle(exc)
        return False

    def _delete_info_dir(self, share_name):
        try:
            shutil.rmtree(
                os.path.join(self.info.info_dir, share_name))
        except:
            pass

    def _get_download_target_fname(self):

        return os.path.join(TEMP_DIR, "chbox_download_target")

    def _check_last_target(self):

        info = self.info

        tgt_root_name = self._get_info(TARGET_ROOT_INFO)
        portal_name = self._get_info(PORTAL_INFO)
        if not tgt_root_name:
            self._set_info(TARGET_ROOT_INFO, info.root)
        if not portal_name:
            self._set_info(PORTAL_INFO, info.portalid)
        if (tgt_root_name != info.root or
            portal_name != info.portalid):
            self._delete_info_dir(info.share_name)
            self._set_info(TARGET_ROOT_INFO, info.root)
            self._set_info(PORTAL_INFO, info.portalid)

    def __write_log_header(self, daytime):

        self._logger.write(LOG_TITLE)
        self._logger.write(LOG_LOCAL_SHARE_TMPL % self.info.share_name)
        self._logger.write(LOG_SERVER_FOLDER_TMPL % self.info.root)
        self._logger.write(LOG_START_DAYTIME_TMPL % \
                               daytime.strftime("%Y/%m/%d %H:%M:%S"))

    def __init_sync_log(self):

        now = dt.datetime.now()
        log_file_path = os.path.join(SYNC_LOG_DIR,
                                     self.info.share_name,
                                     self.__get_sync_start_daytime_str(now))
        self._logger.init_log_param(log_file_path)
        self._logger.set_sync_start_time(now)
        self.__write_log_header(now)

    def __init_download_log(self):

        now = dt.datetime.now()
        log_file_path = os.path.join(SYNC_LOG_DIR,
                                     self.info.share_name,
                                     self.__get_sync_start_daytime_str(now))
        self._logger.init_log_param(log_file_path)
        self._logger.set_sync_start_time(now)
        self.__write_log_header(now)

    def __end_log(self, result):

        try:
            self._logger.set_sync_end_time(dt.datetime.today())
            self._logger.end_log(result)
        except:
            pass

    def get_http_err_id(self, err_data):

        try:
            return json.loads(err_data)[JSON_ERR_ID_KEY]
        except:
            return None

    def make_http_error(self, err, err_data):

        err_stream = StringIO()
        err_stream.write(err_data)
        err_stream.seek(0)
        return urllib2.HTTPError(err.filename, err.code, err.msg,
                          err.hdrs, err_stream)

    def _save_corresponding_event_log(self, log):
        """Arg log is Log class instance."""

        if isinstance(log.exc_arg, NoSpaceError):
            if isinstance(log.exc_arg.detail, urllib2.HTTPError):
                self.__log_event(ERR_LOG_MAP[NO_SPACE_ON_SERVER_ERR_LOG],
                        {"name": self.info.portalid + ":" + self.info.root})

        # skip log
        if log.lv != SKIP_ERROR_LOG_LV:
            return

        if self.current_direction == UPLOADING:
            # upload時はコピー元
            log_arg_prefix = self.info.share_name + "/"
        else:
            log_arg_prefix = self.info.portalid + ":"

        if log.log == FILE_SIZE_OVER_LOG:
            self.__log_event(ERR_LOG_MAP[FILE_TOO_LARGE_ERR_LOG],
                {"name": self.info.share_name + "/" + log.relpath})

        elif log.log == EQUIVALENT_SKIP_LOG:
            self.__log_event(ERR_LOG_MAP[SAME_FILE_NAME_EXISTS_ERR_LOG],
                {"name": self.info.share_name + "/" + log.relpath})

        elif isinstance(log.exc_arg, ForbiddenNameError):
            self.__log_event(ERR_LOG_MAP[FORBIDDEN_FILE_NAME_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

        elif isinstance(log.exc_arg, NameTooLongError):
            self.__log_event(ERR_LOG_MAP[TOO_LONG_FILE_NAME_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

        elif isinstance(log.exc_arg, PermissionError):
            self.__log_event(ERR_LOG_MAP[NO_PERMISSION_ERR_LOG],
                {"name": self.info.portalid + ":" + log.relpath})

        elif isinstance(log.exc_arg, OperationLocked):
            self.__log_event(ERR_LOG_MAP[LOCKED_ERR_LOG],
                {"name": self.info.portalid + ":" + log.relpath})
        else:
            self.__log_event(ERR_LOG_MAP[SKIP_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

    def write_log(self, log):
        """Arg log is Log class instance."""

        self._save_corresponding_event_log(log)
        self._logger.write_log(log)

    def _save_result(self, result):

        result_path = os.path.join(self.info.info_dir,
                                   RESULT_DIR_NAME,
                                   self.info.share_name)

        dir_name = os.path.dirname(result_path)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)

        with open(result_path, "w") as fh:
            fh.write(str(result))

    def _server_session_open(self):

        self._server_tree.session_open()

    def _server_session_close(self):

        try:
            self._server_tree.session_close()
        except:
            pass

    def _set_force_download(self, enable=True):

        try:
            share_name_list = os.listdir(IMMEDIATE_DOWNLOAD_INFO_DIR)
            if self.info.share_name in share_name_list:
                self.is_force_download = True
                os.unlink(os.path.join(IMMEDIATE_DOWNLOAD_INFO_DIR,
                                       self.info.share_name))
        except:
            pass

    def _is_enable_download(self):

        return self.is_force_download or self.info.download_interval

    def __end_handling(self, result):

        try:
            self._server_session_close()
        except:
            pass

        try:
            self._save_result(result)
            self.__end_log(result)
            self.current_direction = None
        except:
            pass

        if self.is_need_retry_sleep:
            time.sleep(RETRY_PERIOD)
            self.is_need_retry_sleep = False

    def _try_exact_first_upload(self):

        self._logger.write(LOG_EXACT_FIRST_UPLOAD)

        try:
            self._try_dst_tree_access_transfer(self._local_tree,
                                               self._server_tree,
                                               do_check_update_fix=True)
        except:
            raise
        else:
            self._post_exact_first_upload()

    def _try_exact_first_download(self):

        self._logger.write(LOG_EXACT_FIRST_DOWNLOAD)

        is_executed = False

        before_uid = os.geteuid()
        before_gid = os.getegid()
        os.setegid(self.info.gid)
        os.seteuid(self.info.uid)

        try:
            last_download_time = self._get_last_time(DOWNLOAD_TIME_INFO)
            if not self._is_exec_download(last_download_time):
                debug(["not exec download"])
                return

            is_executed = True

            self._try_dst_tree_access_transfer(self._server_tree,
                self._local_tree, do_check_update_fix=False)

        except:
            os.seteuid(before_uid)
            os.setegid(before_gid)
            raise
        else:
            os.seteuid(before_uid)
            os.setegid(before_gid)
            if is_executed:
                self._post_exact_first_download()

    def _try_first_download(self):

        self._logger.write(LOG_FIRST_DOWNLOAD)
        super(ChboxSync, self)._try_first_download()

    def _try_normal_upload(self):

        try:
            if self.is_first_upload:
                self._logger.write(LOG_FIRST_UPLOAD)
            else:
                self._logger.write(LOG_NORMAL_UPLOAD)
        except:
            pass

        super(ChboxSync, self)._try_normal_upload()

    def _try_normal_download(self):

        self._logger.write(LOG_NORMAL_DOWNLOAD)
        super(ChboxSync, self)._try_normal_download()

    def try_sync(self):

        result = SUCCESS
        try:
            self.is_need_retry_sleep = False
            self.current_direction = UPLOADING

            self.__init_sync_log()

            # upload
            self._logger.write(LOG_START_UPLOAD)
            self._try_upload()
            self._logger.write(LOG_END_UPLOAD)

            # download
            self._set_force_download()
            if self._is_enable_download():
                self.current_direction = DOWNLOADING
                self._logger.write(LOG_START_DOWNLOAD)
                self._try_download()
                self._logger.write(LOG_END_DOWNLOAD)

        except:
            result = FAILED
            try:
                self._logger.write(LOG_ABORT_UPLOAD + \
                                   lf2crlf(tb.format_exc()), is_need_save=True)
            except:
                pass
            raise
        finally:
            self.__end_handling(result)

    def try_download_immediate(self):

        result = SUCCESS
        try:
            self.is_need_retry_sleep = False
            self.current_direction = DOWNLOADING
            self.is_force_download = True
            self.__init_download_log()
            self._logger.write(LOG_START_DOWNLOAD)
            self._try_download()
            self._logger.write(LOG_END_DOWNLOAD)
        except:
            result = FAILED
            try:
                self._logger.write(LOG_ABORT_DOWNLOAD + \
                                   lf2crlf(tb.format_exc()), is_need_save=True)
            except:
                pass
            raise
        finally:
            self.__end_handling(result)

    def cancel_sync(self):

        try:
            self._logger.write(LOG_CANCEL_SYNC, is_need_save=True)
            self.__end_log(STOPPED)
        except:
            pass
        finally:
            self._server_session_close()


class ChboxCallBack():

    def __init__(self, chbox_sync):

        self._chbox_sync = chbox_sync

    def low_level_except_handler(self, err, target_path=None):

        if isinstance(err, urllib2.HTTPError):

            if UNAUTHORIZED_HTTP_STATUS == err.code:
                raise ServerCommunicationError(detail=err) 

            if not hasattr(err, "read"):
                raise err

            err_data = err.read()
            err_id = self._chbox_sync.get_http_err_id(err_data)
            new_err = self._chbox_sync.make_http_error(err, err_data)
            if err_id in errid_except_dict:

                # ルートディレクトリが存在しないなら
                if target_path is not None:
                    if self._chbox_sync.info.root == target_path:

                        if errid_except_dict[err_id] == NoEntryError:
                            raise NoRootEntryError(detail=new_err)

                        if errid_except_dict[err_id] == ForbiddenNameError:
                            raise NoRootEntryError(detail=new_err)

                raise errid_except_dict[err_id](detail=new_err)
            else:
                raise err

        elif (isinstance(err, UnicodeError) or
              isinstance(err, UnicodeDecodeError) or
              isinstance(err, UnicodeEncodeError)):
            raise ForbiddenNameError(detail=err)

        elif isinstance(err, urllib2.URLError):
            raise err

        raise err

    def log_callback(self, log):
        """Arg log is Log class instance."""

        self._chbox_sync.write_log(log)


class ChboxSyncMgr(SyncMgr):
    """
    設定ファイルの仕様を早く決める必要あり
    共有にログを残す話もあるのでそれも考慮する必要ある
    """

    INFO_DIR = SYNC_INFO_DIR

    SHARE_CONF_SERVICE_ENABLE_PATH = \
        "sharelist/share/shareservice/chbox[@enable='true']/../.."

    ANNOUNCE_CAT = Cat.CHBOX

    NASCMD_FOR_LOGGING = NasCmd(Results, logger_name="CHBOX",
        log=LOG_DIR + "/chbox.log", log_level=logging.DEBUG)

    class ChboxSyncInfo(EconomySyncInfo):

        def __init__(self, share_name, share_abs_path,
                     uid, gid, download_interval,
                     portalid, login_id, login_passwd,
                     root, uploadmaxfilesize,
                     proxy_host, proxy_port,
                     info_dir):
            super(self.__class__, self).__init__(share_name,
                                                 share_abs_path,
                                                 uid,
                                                 gid,
                                                 download_interval)
            self.portalid = portalid
            self.login_id = login_id
            self.login_passwd = login_passwd
            self.root = root
            self.uploadmaxfilesize = uploadmaxfilesize * 1024 * 1024  # MB
            self.proxy_host = proxy_host
            self.proxy_port = proxy_port
            self.info_dir = info_dir

    def _load_share_list(self):

        SHARE_XPATH = "sharelist/share"
        CHBOXSYNC_XPATH = "shareservice/chboxsync"
        PATH_XPATH_TMPL = "sharelist/share[name='%s']/path"

        shares_node = send_request(FAL_SHARES_PATH, ReqTypes.LOAD)
        chbox_node = send_request(FAL_CHBOX_PATH, ReqTypes.LOAD)

        share_list = []
        try:
            for tmp_share_node in chbox_node.get_node_list(SHARE_XPATH):
                if not tmp_share_node.get_node(CHBOXSYNC_XPATH).is_enabled():
                    continue
                name = tmp_share_node.get_value(NAME_NAME)
                share_node = tmp_share_node.get_clone()
                share_node.append_child(
                    shares_node.get_node(PATH_XPATH_TMPL % name))
                share_list.append(share_node)
            return share_list
        except NotFoundError:
            return []

    def _get_share_user_group_id(self, share):

        nobody_uid = pwd.getpwnam("nobody")[2]
        nogroup_gid = grp.getgrnam("nogroup")[2]

        uid = share.get_value(UID_NAME, nobody_uid)
        gid = share.get_value(GID_NAME, nogroup_gid)

        return (int(uid), int(gid))

    def _get_sync_info(self, share):

        chbox_node = share.get_node(CHBOXSYNC_PATH)

        uid, gid = self._get_share_user_group_id(share)

        proxy_host, proxy_port = self._get_proxy(True, REST_BASE_HOST)

        return self.ChboxSyncInfo(
            share.get_node(NAME_NAME).get_text(),
            share.get_node(PATH_NAME).get_text(),
            uid,
            gid,
            int(chbox_node.get_node(DOWNLOADINTERVAL_NAME).get_text()),
            chbox_node.get_node(PORTALID_NAME).get_text(do_strip=False),
            chbox_node.get_node(LOGINID_NAME).get_text(do_strip=False),
            chbox_node.get_node(LOINPASSWD_NAME).get_text(do_strip=False),
            chbox_node.get_node(ROOT_PATH_XPATH).get_text(
                do_strip=False).rstrip(CHBOX_SEP),
            int(chbox_node.get_node(
                UPLOAD_MAX_SIZE_NAME).get_text(do_strip=False)),
            proxy_host,
            proxy_port,
            info_dir=self.INFO_DIR)

    def _make_sync_dict(self, share):

        info = self._get_sync_info(share)
        return {info.share_name: ChboxSync(info)}

    def _save_result(self, result, share_name):

        result_path = os.path.join(self.INFO_DIR, RESULT_DIR_NAME, share_name)
        dir_name = os.path.dirname(result_path)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)
        with open(result_path, "w") as fh:
            fh.write(str(result))

    def _save_fail_result(self, fail_share_list):

        for share in fail_share_list:
            try:
                self._save_result(FAILED, share)
            except:
                pass

    def _get_sync_dict(self):

        fail_list = []
        sync_dict = {}

        for share in self._load_share_list():
            try:
                sync_dict.update(self._make_sync_dict(share))
            except:
                debug(do_tb=True)
                try:
                    fail_list.append(share.get_node(NAME_NAME).get_text())
                except:
                    pass
                continue

        self._delete_disable_info_dir(sync_dict)
        self._save_fail_result(fail_list)
        return sync_dict

    def _get_proxy(self, is_secure, endpoint):
        """
        is_secureはいらない
        """

        if is_secure:
            url = "https://" + endpoint + "/"
        else:
            url = "http://" + endpoint + "/"

        proxy_info = get_proxy(url)

        if proxy_info:
            return tuple(proxy_info.split(":"))
        else:
            return (None, None)

    def _reset_proxy_info(self):
        for share in self.sync_dict:
            sync = self.sync_dict[share]
            host, port = self._get_proxy(True, REST_BASE_HOST)
            sync.reset_proxy_info(host, port)

    def _get_info(self, share_name, info):
        try:
            with open(os.path.join(self.INFO_DIR,
                                   share_name,
                                   info),
                      "rb") as file:
                return file.read()
            return False
        except:
            return False

    def _clear_announce(self):
        clear_announce("all", self.ANNOUNCE_CAT)

    def download_immediate(self, tgt_share_name_list):

        try:
            debug([self.__class__.__name__, dt.datetime.now()],
                  False, fmt="/// %s CloudSyncStart /// %s\n")

            is_cur_err = False

            for share_name, sync in self.sync_dict.iteritems():

                if share_name not in tgt_share_name_list:
                    continue

                try:
                    sync.try_download_immediate()

                except Exception:
                    debug(do_tb=True)

                    if not sync.is_disable:
                        is_cur_err = True

                    self._reset_proxy_info()

                finally:
                    debug([share_name, dt.datetime.now()],
                          False, fmt="+++ %s ShareSyncEnd +++ %s\n")

            if self.is_last_error and not is_cur_err:
                self._clear_announce()

            self.is_last_error = is_cur_err

        finally:
            debug([self.__class__.__name__, dt.datetime.now()],
                  False, fmt="/// %s CloudSyncEnd /// %s\n")

    def sigterm_handler(self, signum, frame):

        for share, sync in self.sync_dict.iteritems():
            try:
                sync.cancel_sync()
            except:
                pass

        os._exit(0)
