#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

import re
import os
import time

from iobase.xmldata import XmlNode
from ionas.request import Request, RequestData, Entry
from ionas.exception import InvalidFileSystemError
from fal_info import (ReqTypes, Results,
    FAL_PROXY_PATH, FAL_PROXY_REQ_BASE_XML,
    FAL_ANNOUNCE_PATH, FAL_EVENTLOG_PATH, FAL_SHARES_PATH, FAL_VOLCTRL_PATH)
from product_conf import (STOR_IF_LIST, STOR_IF_TO_VOL, TEMP_DIR)


ARGLIST_NAME = "arglist"
ARG_NAME = "arg"
LOG_NAME = "log"
URL_NAME = "url"
SERVERLIST_NAME = "serverlist"
SERVER_NAME = "server"
NAME_PATH = SERVER_NAME + "/host"
PORT_PATH = SERVER_NAME + "/port"
HOST_NAME = "host"
PORT_NAME = "port"
NAME_ATTR = "name"
PROXY_DIRECT = "DIRECT"
PROC_MOUNTS_DIR = "/proc/mounts"
VALID_FS_LIST = ("ext2", "ext3", "ext4", "xfs")
MOUNTS_OPT_SPLIT = ","
RW_OPT = "rw"
MOUNT_ROOT = "/mnt"
EXTERNAL_SHARE_DICT = {"usb1": MOUNT_ROOT + "/usb1",
                        "usb2": MOUNT_ROOT + "/usb2",
                        "usb3": MOUNT_ROOT + "/usb3",
                        "usb4": MOUNT_ROOT + "/usb4"}

DOWNLOAD_TMP_DIR = MOUNT_ROOT + "/sataraid1/nasdsync_download_tmp"

CLEAR_ANNOUNCE_BASE_REQ_XML = \
  '<announce>' + \
  ' <option level=\"%s\" category=\"%s\"></option>' + \
  '</announce>'

BOOT_COMPLETE_FILE_PATH = TEMP_DIR + "/boot_complete"

MOUNT_TIMEOUT = 10 * 60 * 60  # 10 min


class RequestError(Exception):
    """Request error"""

    pass

class TimeOutError(Exception):
    """TimeOut error"""

    pass

def __is_rw_opt_set(options):
    """Check if rw option is set."""

    for opt in options.split(MOUNTS_OPT_SPLIT):
        if opt == RW_OPT:
            return True
    return False


def __is_mount_point_available(mount_point):
    """Check if mount_point is available."""

    mount_re = re.compile("[^ ]+ " + mount_point + " ([^ ]+) ([^ ]+).*")
    for line in open(PROC_MOUNTS_DIR):
        match = mount_re.match(line)
        if not match:
            continue

        if not match.group(1) in VALID_FS_LIST:
            raise InvalidFileSystemError

        return __is_rw_opt_set(match.group(2))

    return False


def get_proxy(url):
    """Try to get proxy and return it."""

    entry = Entry(ReqTypes.QUERY,
                    FAL_PROXY_REQ_BASE_XML[ReqTypes.QUERY])

    entry.get().get_node(URL_NAME).set_text(url)
    request_data = RequestData(entry)
    target = Request(FAL_PROXY_PATH)
    result_entry = target.request(request_data).get_entrylist().get_entry()
    if result_entry.get_result() != Results.OK:
        return ()

    serverlist = result_entry.get().get_node(SERVERLIST_NAME)
    for proxy in serverlist.get_node_list(SERVER_NAME):
        host = proxy.get_node(HOST_NAME).get_text()
        port = proxy.get_node(PORT_NAME).get_text()

        if not host or host == PROXY_DIRECT:
            continue

        proxy_server = host
        if port:
            proxy_server += ':' + port
        return proxy_server

    return ""


def is_share_available(share, share_abs_path):
    """Check if local share folder is available."""

    if not os.path.isdir(share_abs_path):
        return False

    if not share in EXTERNAL_SHARE_DICT:
        return True

    return __is_mount_point_available(share_abs_path)


def get_tmp_path(share_abs_path):
    if share_abs_path in EXTERNAL_SHARE_DICT.values():
        return share_abs_path
    return DOWNLOAD_TMP_DIR


def send_request(cmd, type, req_xml=""):

    response_entry = Request(cmd).request(Entry(type, req_xml)).get_entry()
    if response_entry.get_result() != Results.OK:
        raise RequestError
    return response_entry.get()


def save_event_log(msg_code, arg_dict={}):
    """save event log."""

    EVENTLOG_XML_TMPL = '<eventlog><log><code>%s</code></log></eventlog>'

    eventlog_node = XmlNode(EVENTLOG_XML_TMPL % msg_code)

    arglist_node = XmlNode(ARGLIST_NAME)
    for arg_name, arg_value in arg_dict.iteritems():
        arglist_node.append_child(ARG_NAME, arg_value, {NAME_ATTR: arg_name})
        eventlog_node.get_node(LOG_NAME).append_child(arglist_node)

    return send_request(FAL_EVENTLOG_PATH, ReqTypes.ADD, eventlog_node)

def clear_announce(level="1", category=""):
    """Clear announe to level and category.
       "level" can be set "1" or "2" or "all".
       "category" set "" is all category.
    """
    return send_request(FAL_ANNOUNCE_PATH, ReqTypes.CLEAR,
                        CLEAR_ANNOUNCE_BASE_REQ_XML % (level, category))

def get_sync_log_dir_path():

    LOG_SHARE_NAME = "LAN DISK Log"
    shares_node = send_request(FAL_SHARES_PATH, ReqTypes.LOAD)
    return shares_node.get_value(
               "sharelist/share[name='%s']/path" % LOG_SHARE_NAME, "")


def is_boot_complete():

    if os.path.exists(BOOT_COMPLETE_FILE_PATH):
        return True
    return False


def wait_for_internal_volume_mounted(timeout_sec=MOUNT_TIMEOUT):

    start = time.time()
    while True:

        if is_boot_complete():
            break

        time.sleep(5)

        now = time.time()
        if now - start > timeout_sec:
            raise TimeOutError
